/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>

#include "AMRLevelGodunovFactory.H"
#include "AMRLevelGodunov.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "ParmParse.H"
#include "HDF5.H"
#include "parstream.H"

#include "RampIBC.H"
#include "ExplosionIBC.H"
#include "GaussianIBC.H"
#include "SineIBC.H"
#include "WaveIBC.H"

#define PROBLEM_RAMP      0
#define PROBLEM_EXPLOSION 1
#define PROBLEM_GAUSSIAN  2
#define PROBLEM_SINE      3
#define PROBLEM_WAVE      4

#define PRESSURE_ISENTROPIC 0
#define PRESSURE_CONSTANT   1

// amrGodunov is a function (as opposed to inline in main()) to get
// around mpi scoping problems

void amrGodunov();

void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_max_level,
                     int                   a_max_grid_size,
                     int                   a_block_factor,
                     int                   a_verbosity,
                     std::string           a_grid_file);

void dumpmemoryatexit();

int main(int a_argc, char* a_argv[])
{
#ifdef MPI
  MPI_Init(&a_argc,&a_argv);
  setChomboMPIErrorHandler();
#endif

  // read inputs
  char* in_file = NULL;

  if(a_argc > 1)
  {
    in_file = a_argv[1];
  }
  else
  {
    pout() << "Usage:  amrGodunov...ex <inputfile>" << endl;
    pout() << "No input file specified" << endl;
    return -1;
  }

  ParmParse pp(a_argc-2,a_argv+2,NULL,in_file);

  // run amrGodunov
  amrGodunov();

#ifdef MPI
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

void amrGodunov()
{
  // read inputs
  ParmParse ppgodunov("godunov");

  int problem = -1;
  std::string problemString;

  if (ppgodunov.contains("problem"))
  {
    ppgodunov.query("problem",problemString);
    if (problemString == "ramp")
    {
      problem = PROBLEM_RAMP;
    }
    else if (problemString == "explosion")
    {
      problem = PROBLEM_EXPLOSION;
    }
    else if (problemString == "gaussian")
    {
      problem = PROBLEM_GAUSSIAN;
    }
    else if (problemString == "sine")
    {
      problem = PROBLEM_SINE;
    }
    else if (problemString == "wave")
    {
      problem = PROBLEM_WAVE;
    }

    if (problem == -1)
    {
      pout() << "Invalid problem, \"" << problemString << "\", specified in input file" << endl << endl; 
      return;
    }
  }
  else
  {
      pout() << "\"godunov.problem\" not specified in input file" << endl << endl; 
      return;
  }

  int verbosity = 0;
  ppgodunov.query("verbosity",verbosity);
  assert(verbosity >= 0);

  // For all gas dynamics
  Real gamma = 1.4;
  ppgodunov.query("gamma",gamma);

  // For "ramp"
  Real alpha = 30.0;
  Real xcorner = 0.1;

  // For "ramp" and "explosion"
  Real ms = 10.0;

  // For "explosion" and "gaussian"
  Real size = 0.25;

  // For "explosion", "gaussian", "sine", and "wave"
  vector<Real> centerpp(SpaceDim,0.5);
  RealVect center;
  vector<Real> velocitypp(SpaceDim,0.0);
  RealVect velocity;

  // For "gaussian", "sine", and "wave"
  Real ambientDensity = 1.4;
  Real deltaDensity = 0.014;
  int pressure = -1;
  std::string pressureString;

  // For "wave"
  vector<int> waveNumberpp(SpaceDim,0);
  waveNumberpp[0] = 1;
  IntVect waveNumber;

  switch (problem)
  {
    case PROBLEM_RAMP:
    {
      ppgodunov.query("angle_deg",alpha);
      ppgodunov.query("shock_mach",ms);
      ppgodunov.query("xcorner",xcorner);
      break;
    }

  case PROBLEM_EXPLOSION:
    {
      ppgodunov.query("shock_mach",ms);

      ppgodunov.queryarr("center",centerpp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

      ppgodunov.query("size",size);

      ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

      break;
    }

  case PROBLEM_GAUSSIAN:
    {
      ppgodunov.query("ambient_density",ambientDensity);
      ppgodunov.query("delta_density",deltaDensity);

      if (ppgodunov.contains("pressure"))
      {
        ppgodunov.query("pressure",pressureString);
        if (pressureString == "isentropic")
        {
          pressure = PRESSURE_ISENTROPIC;
        }
        else if (pressureString == "constant")
        {
          pressure = PRESSURE_CONSTANT;
        }

        if (pressure == -1)
        {
          pout() << "Invalid pressure, \"" << pressureString << "\", specified in input file" << endl << endl; 
          return;
        }
      }
      else
      {
        pressure = PRESSURE_ISENTROPIC;
      }

      ppgodunov.queryarr("center",centerpp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

      ppgodunov.query("size",size);

      ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

      break;
    }
    
  case PROBLEM_SINE:
    {
      ppgodunov.query("ambient_density",ambientDensity);
      ppgodunov.query("delta_density",deltaDensity);

      if (ppgodunov.contains("pressure"))
      {
        ppgodunov.query("pressure",pressureString);
        if (pressureString == "isentropic")
        {
          pressure = PRESSURE_ISENTROPIC;
        }
        else if (pressureString == "constant")
        {
          pressure = PRESSURE_CONSTANT;
        }

        if (pressure == -1)
        {
          pout() << "Invalid pressure, \"" << pressureString << "\", specified in input file" << endl << endl; 
          return;
        }
      }
      else
      {
        pressure = PRESSURE_ISENTROPIC;
      }

      ppgodunov.queryarr("center",centerpp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

      ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

      break;
    }
    
  case PROBLEM_WAVE:
    {
      ppgodunov.query("ambient_density",ambientDensity);
      ppgodunov.query("delta_density",deltaDensity);

      if (ppgodunov.contains("pressure"))
      {
        ppgodunov.query("pressure",pressureString);
        if (pressureString == "isentropic")
        {
          pressure = PRESSURE_ISENTROPIC;
        }
        else if (pressureString == "constant")
        {
          pressure = PRESSURE_CONSTANT;
        }

        if (pressure == -1)
        {
          pout() << "Invalid pressure, \"" << pressureString << "\", specified in input file" << endl << endl; 
          return;
        }
      }
      else
      {
        pressure = PRESSURE_ISENTROPIC;
      }

      ppgodunov.queryarr("wave_number",waveNumberpp,0,SpaceDim);
      int norm2 = 0;
      for (int i = 0; i < SpaceDim; i++)
      {
        waveNumber[i] = waveNumberpp[i];
        norm2 += waveNumber[i]*waveNumber[i];
      }

      if (norm2 == 0)
      {
        pout() << "One component of the wave number must be non-zero" << endl << endl;
        return;
      }

      ppgodunov.queryarr("center",centerpp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

      ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
      for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

      break;
    }
    
    default:
    {
      MayDay::Error("amrGodunov \"problem\" (used for initial and boundary conditions) is invalid");
      break;
    }
  }

  int nstop = 0;
  ppgodunov.query("max_step",nstop);

  Real stopTime = 0.0;
  ppgodunov.query("max_time",stopTime);

  vector<int> num_cells(SpaceDim);
  for (int i = 0; i < SpaceDim; ++i) num_cells[i]=0;
  ppgodunov.queryarr("num_cells",num_cells,0,SpaceDim);

  assert(D_TERM(   (num_cells[0] > 0),
                && (num_cells[1] > 0),
                && (num_cells[2] > 0)));
  assert(D_TERM(   (num_cells[0] % 2 == 0),
                && (num_cells[1] % 2 == 0),
                && (num_cells[2] % 2 == 0)));

  vector<int> is_periodica(SpaceDim,0);
  bool is_periodic[SpaceDim];
  
  ppgodunov.queryarr("is_periodic",is_periodica,0,SpaceDim);
  // convert periodic from int->bool
  for (int dim=0; dim<SpaceDim; dim++) 
    {
      is_periodic[dim] = (is_periodica[dim] == 1);
      if (is_periodic[dim] && verbosity >= 2 && procID() == 0)
        pout() << "Using Periodic BCs in direction: " << dim << endl;
    }

  int max_level = 0;
  ppgodunov.query("max_level",max_level);

  int num_read_levels = Max(max_level,1);
  std::vector<int> ref_ratios; // (num_read_levels,1);
  // note this requires a ref_ratio to be defined for the 
  // finest level (even though it will never be used)
  ppgodunov.queryarr("ref_ratio",ref_ratios,0,num_read_levels+1);

  std::vector<int> regrid_intervals; // (num_read_levels,1);
  ppgodunov.queryarr("regrid_interval",regrid_intervals,0,num_read_levels);

  Real refine_thresh = 0.3;
  ppgodunov.query ("refine_thresh",refine_thresh);

  int block_factor = 1;
  ppgodunov.query("block_factor",block_factor);

  int tag_buffer_size = 3;
  ppgodunov.query("tag_buffer_size",tag_buffer_size);

  int max_grid_size = 32;
  ppgodunov.query("max_grid_size",max_grid_size);

  Real fill_ratio = 0.75;
  ppgodunov.query("fill_ratio",fill_ratio);

  int checkpoint_interval = 0;
  ppgodunov.query("checkpoint_interval",checkpoint_interval);

  int plot_interval = 0;
  ppgodunov.query("plot_interval",plot_interval);

  Real cfl = 0.8;
  ppgodunov.query("cfl",cfl);

  Real initial_cfl = 0.1;
  ppgodunov.query("initial_cfl",initial_cfl);

  Real fixed_dt = -1;
  ppgodunov.query("fixed_dt",fixed_dt);

  Real max_dt_growth = 1.1;
  ppgodunov.query("max_dt_growth",max_dt_growth);

  Real dt_tolerance_factor = 1.1;
  ppgodunov.query("dt_tolerance_factor",dt_tolerance_factor);

  if ( verbosity >= 2 )
  {
    pout() << "problem = " << problemString << endl;
    pout() << "gamma = " << gamma << endl;

    // display inputs
    switch (problem)
    {
      case PROBLEM_RAMP:
      {
        pout() << "alpha = " << alpha << endl;
        pout() << "shock mach = " << ms << endl;
        pout() << "xcorner = " << xcorner << endl;
        break;
      }

    case PROBLEM_EXPLOSION:
      {
        pout() << "shock mach = " << ms << endl;
        pout() << "center = " << D_TERM(center[0] << "  " <<,
                                        center[1] << "  " <<,
                                        center[2] << ) endl;
        pout() << "size = " << size << endl;
        pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                          velocity[1] << "  " <<,
                                          velocity[2] << ) endl;
        break;
      }

    case PROBLEM_GAUSSIAN:
      {
        pout() << "ambient density = " << ambientDensity << endl;
        pout() << "delta density = " << deltaDensity << endl;
        pout() << "pressure = " << pressureString << endl;
        pout() << "center = " << D_TERM(center[0] << "  " <<,
                                        center[1] << "  " <<,
                                        center[2] << ) endl;
        pout() << "size = " << size << endl;
        pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                          velocity[1] << "  " <<,
                                          velocity[2] << ) endl;
        break;
      }

    case PROBLEM_SINE:
      {
        pout() << "ambient density = " << ambientDensity << endl;
        pout() << "delta density = " << deltaDensity << endl;
        pout() << "pressure = " << pressureString << endl;
        pout() << "center = " << D_TERM(center[0] << "  " <<,
                                        center[1] << "  " <<,
                                        center[2] << ) endl;
        pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                          velocity[1] << "  " <<,
                                          velocity[2] << ) endl;
        break;
      }
      
    case PROBLEM_WAVE:
      {
        pout() << "ambient density = " << ambientDensity << endl;
        pout() << "delta density = " << deltaDensity << endl;
        pout() << "pressure = " << pressureString << endl;
        pout() << "wave number = " << D_TERM(waveNumber[0] << "  " <<,
                                             waveNumber[1] << "  " <<,
                                             waveNumber[2] << ) endl;
        pout() << "center = " << D_TERM(center[0] << "  " <<,
                                        center[1] << "  " <<,
                                        center[2] << ) endl;
        pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                          velocity[1] << "  " <<,
                                          velocity[2] << ) endl;
        break;
      }
      
      default:
      {
        MayDay::Error("amrGodunov \"problem\" (used for initial and boundary conditions) is invalid");
        break;
      }
    }

    pout() << "maximum step = " << nstop << endl;
    pout() << "maximum time = " << stopTime << endl;

    pout() << "number of cells = " << D_TERM(num_cells[0] << "  " <<,
                                             num_cells[1] << "  " <<,
                                             num_cells[2] << ) endl;

    pout() << "maximum level = " << max_level << endl;

    pout() << "refinement ratio = ";
    for (int i = 0; i < ref_ratios.size(); ++i) pout() << ref_ratios[i] << " ";
    pout() << endl;

    pout() << "regrid interval = ";
    for (int i = 0; i < regrid_intervals.size(); ++i) pout() << regrid_intervals[i] << " ";
    pout() << endl;

    pout() << "refinement threshold = " << refine_thresh << endl;

    pout() << "blocking factor = " << block_factor << endl;
    pout() << "max grid size = " << max_grid_size << endl;
    pout() << "fill ratio = " << fill_ratio << endl;

    pout() << "checkpoint interval = " << checkpoint_interval << endl;
    pout() << "plot interval = " << plot_interval << endl;
    pout() << "CFL = " << cfl << endl;
    pout() << "initial CFL = " << initial_cfl << endl;
    if (fixed_dt > 0)
    {
      pout() << "fixed dt = " << fixed_dt << endl;
    }
    pout() << "maximum dt growth = " << max_dt_growth << endl;
    pout() << "dt tolerance factor = " << dt_tolerance_factor << endl;
  }

  // create objects
  RefCountedPtr<PhysIBC> ibc;

  switch (problem)
  {
    case PROBLEM_RAMP:
    {
      RampIBC *rampibc = new RampIBC;
      rampibc->setFortranCommon(gamma,
                                alpha,
                                ms,
                                xcorner);
      ibc = rampibc;
      break;
    }

  case PROBLEM_EXPLOSION:
    {
      ExplosionIBC * explosionibc = new ExplosionIBC;
      explosionibc->setFortranCommon(gamma,
                                     ms,
                                     center,
                                     size,
                                     velocity);
      ibc = explosionibc;
      break;
    }

  case PROBLEM_GAUSSIAN:
    {
      GaussianIBC * gaussianibc = new GaussianIBC;
      gaussianibc->setFortranCommon(gamma,
                                    ambientDensity,
                                    deltaDensity,
                                    pressure,
                                    center,
                                    size,
                                    velocity);
      ibc = gaussianibc;
      break;
    }

  case PROBLEM_SINE:
    {
      SineIBC * sineibc = new SineIBC;
      sineibc->setFortranCommon(gamma,
                                ambientDensity,
                                deltaDensity,
                                pressure,
                                center,
                                velocity);
      ibc = sineibc;
      break;
    }
    
  case PROBLEM_WAVE:
    {
      WaveIBC * waveibc = new WaveIBC;
      waveibc->setFortranCommon(gamma,
                                ambientDensity,
                                deltaDensity,
                                pressure,
                                waveNumber,
                                center,
                                velocity);
      ibc = waveibc;
      break;
    }
    
    default:
    {
      MayDay::Error("amrGodunov \"problem\" (used for initial and boundary conditions) is invalid");
      break;
    }
  }

  ProblemDomain prob_domain (IntVect::Zero,
                             IntVect(D_DECL(num_cells[0]-1,
                                            num_cells[1]-1,
                                            num_cells[2]-1)),
                             is_periodic);

  AMRLevelGodunovFactory amrg_fact;

  amrg_fact.CFL(cfl);
  amrg_fact.physIBC(ibc);
  amrg_fact.refinementThreshold(refine_thresh);
  amrg_fact.tagBufferSize(tag_buffer_size);

  // 2.0 multiplier is here because of double sweep in LevelGodunov
  amrg_fact.initialDtMultiplier(2.0*initial_cfl);

  AMR amr;

  amr.define(max_level,ref_ratios,prob_domain,&amrg_fact);

  if (fixed_dt > 0)
  {
    amr.fixedDt(fixed_dt);
  }

  // set grid generation parameters
  amr.maxGridSize(max_grid_size);
  amr.blockFactor(block_factor);
  amr.fillRatio(fill_ratio);

  // the hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // set output parameters
  amr.checkpointInterval(checkpoint_interval);
  amr.plotInterval(plot_interval);
  amr.regridIntervals(regrid_intervals);
  amr.maxDtGrow(max_dt_growth);
  amr.dtToleranceFactor(dt_tolerance_factor);

  if (ppgodunov.contains("plot_prefix"))
  {
    std::string prefix;
    ppgodunov.query("plot_prefix",prefix);
    amr.plotPrefix(prefix);
  }

  if (ppgodunov.contains("chk_prefix"))
  {
    std::string prefix;
    ppgodunov.query("chk_prefix",prefix);
    amr.checkpointPrefix(prefix);
  }

  amr.verbosity(verbosity);

  if (!ppgodunov.contains("restart_file"))
  {
    if (!ppgodunov.contains("fixed_hierarchy"))
    {
      // initialize from scratch for AMR run
      // initialize hierarchy of levels
      amr.setupForNewAMRRun();
    }
    else
    {
      std::string grid_file;
      ppgodunov.query("fixed_hierarchy",grid_file);

      // initialize from a list of grids in "grid_file"
      Vector<Vector<Box> > amrGrids(max_level+1);
      setupFixedGrids(amrGrids,
                      prob_domain,
                      max_level,
                      max_grid_size,
                      block_factor,
                      verbosity,
                      grid_file);
      amr.setupForFixedHierarchyRun(amrGrids,1);
    }
  }
  else
  {
    std::string restart_file;
    ppgodunov.query("restart_file",restart_file);

#ifdef HDF5
      HDF5Handle handle(restart_file,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }

  // run 
  amr.run(stopTime,nstop);

  // output last pltfile and statistics
  amr.conclude();
}

void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_max_level,
                     int                   a_max_grid_size,
                     int                   a_block_factor,
                     int                   a_verbosity,
                     std::string           a_grid_file)
{
  if (procID() == uniqueProc(SerialTask::compute))
  {
    a_amrGrids.push_back(Vector<Box>(1,a_domain.domainBox()));

    // read in predefined grids
    ifstream is(a_grid_file.c_str(), ios::in);

    if (is.fail())
    {
      MayDay::Error("Cannot open grids file");
    }
    
    // format of file -- number of levels, then for each level starting
    // with level 1, number of grids on level, list of boxes

    int in_numLevels;
    is >> in_numLevels;

    assert (in_numLevels <= a_max_level+1);

    if (a_verbosity >= 3)
    {
      pout() << "numLevels = " << in_numLevels << endl;
    }

    while (is.get() != '\n');

    a_amrGrids.resize(in_numLevels);

    // check to see if coarsest level needs to be broken up
    domainSplit(a_domain,a_amrGrids[0],a_max_grid_size,a_block_factor);
    
    if (a_verbosity >= 3)
    {
      pout() << "level 0: ";
      for (int n = 0; n < a_amrGrids[0].size(); n++)
      {
        pout() << a_amrGrids[0][0] << endl;
      }
    }

    // now loop over levels, starting with level 1
    int ngrid;
    for (int lev = 1; lev < in_numLevels; lev++)
    {
      is >> ngrid;

      if (a_verbosity >= 3)
      {
        pout() << "level " << lev << " numGrids = " << ngrid << endl;
        pout() << "Grids: ";
      }

      while (is.get() != '\n');

      a_amrGrids[lev].resize(ngrid);

      for (int i = 0; i < ngrid; i++)
      {
        Box bx;
        is >> bx;

        while (is.get() != '\n');

        // quick check on box size
        Box bxRef(bx);

        if (bxRef.longside() > a_max_grid_size)
        {
          pout() << "Grid " << bx << " too large" << endl;
          MayDay::Error();
        }

        if (a_verbosity >= 3) {
          pout() << bx << endl;
        }

        a_amrGrids[lev][i] = bx;
      } // end loop over boxes on this level
    } // end loop over levels
  }

  broadcast(a_amrGrids,uniqueProc(SerialTask::compute));
}
