/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
#include <cmath>
#include <cstdio>

#include "parstream.H"

#include "Box.H"
#include "Vector.H"
#include "IntVectSet.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LayoutIterator.H"
#include "HDF5.H"
#include "SPMD.H"
#include "LoadBalance.H"
#include "LevelFluxRegister.H"
#include "ProblemDomain.H"

#include "PhysIBC.H"
#include "GodunovF_F.H"
#include "BoxIterator.H"

#include "AMRLevel.H"
#include "AMRLevelGodunov.H"

AMRLevelGodunov::AMRLevelGodunov()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov default constructor" << endl;
  }

  m_cfl = 0.8;
  m_refine_thresh = 0.2;
  m_initial_dt_multiplier = 0.1;

  m_num_comps = 2 + SpaceDim;

  m_state_names.push_back("density");
  m_state_names.push_back("x-momentum");
  if (SpaceDim >= 2)
  {
    m_state_names.push_back("y-momentum");

    if (SpaceDim >= 3)
    {
      m_state_names.push_back("z-momentum");
    }
  }
  m_state_names.push_back("energy-density");

  m_phys_ibc_ptr = NULL;
}

// destructor
AMRLevelGodunov::~AMRLevelGodunov()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov destructor" << endl;
  }
}

void AMRLevelGodunov::define(AMRLevel*  a_coarser_level_ptr,
                             const Box& a_problem_domain,
                             int        a_level,
                             int        a_ref_ratio)
{
  ProblemDomain physdomain(a_problem_domain);

  define(a_coarser_level_ptr,physdomain,a_level,a_ref_ratio);
}

void AMRLevelGodunov::define(AMRLevel*  a_coarser_level_ptr,
                             const ProblemDomain& a_problem_domain,
                             int        a_level,
                             int        a_ref_ratio)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::define " << a_level << endl;
  }

  AMRLevel::define(a_coarser_level_ptr,
                   a_problem_domain,
                   a_level,
                   a_ref_ratio);

  m_dx = 1.0 / a_problem_domain.domainBox().longside();

  if (a_coarser_level_ptr != NULL)
  {
    AMRLevelGodunov* amrg_ptr = dynamic_cast <AMRLevelGodunov*> (a_coarser_level_ptr);
    if (amrg_ptr != NULL)
    {
      m_cfl = amrg_ptr->m_cfl;
      m_refine_thresh = amrg_ptr->m_refine_thresh;
      m_tag_buffer_size = amrg_ptr->m_tag_buffer_size;
    }
    else
    {
      MayDay::Error("AMRLevelGodunov::define: a_coarser_level_ptr is not castable to AMRLevelGodunov*");
    }
  }
}

// advance conserved state by one timestep
Real AMRLevelGodunov::advance()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::advance level " << m_level << " at time " << m_time << endl;
  }

  m_state_new.copyTo(m_state_new.interval(),
                     m_state_old,
                     m_state_old.interval());

  Real new_dt = 0.0;

  // finer and coarser levels exist
  if ((m_coarser_level_ptr != NULL) && (m_finer_level_ptr != NULL))
  {
    AMRLevelGodunov* amrg_coarser_ptr = dynamic_cast <AMRLevelGodunov*> (m_coarser_level_ptr);
    AMRLevelGodunov* amrg_finer_ptr   = dynamic_cast <AMRLevelGodunov*> (m_finer_level_ptr);
    if ((amrg_coarser_ptr != NULL) && (amrg_finer_ptr != NULL))
    {
      Real dtcnew = amrg_coarser_ptr->m_time;
      Real dtcold = dtcnew - amrg_coarser_ptr->m_dt;

      new_dt = m_levelgodunov.step(m_state_new,
                                   m_flux_register,
                                   amrg_coarser_ptr->m_flux_register,
                                   amrg_coarser_ptr->m_state_old,
                                   amrg_coarser_ptr->m_state_new,
                                   m_time,
                                   dtcold,
                                   dtcnew,
                                   m_dt);
    }
    else
    {
      MayDay::Error("in AMRLevelGodunov::advance: m_coarser_level_ptr or m_finer_level_ptr is not castable to AMRLevelGodunov*");
    }
  }
  // coarser level exists but not finer
  else if (m_coarser_level_ptr != NULL)
  {
    AMRLevelGodunov* amrg_coarser_ptr = dynamic_cast <AMRLevelGodunov*> (m_coarser_level_ptr);
    if (amrg_coarser_ptr != NULL)
    {
      Real dtcnew = amrg_coarser_ptr->m_time;
      Real dtcold = dtcnew - amrg_coarser_ptr->m_dt;

      new_dt = m_levelgodunov.step(m_state_new,
                                   m_flux_register,
                                   amrg_coarser_ptr->m_flux_register,
                                   amrg_coarser_ptr->m_state_old,
                                   amrg_coarser_ptr->m_state_new,
                                   m_time,
                                   dtcold,
                                   dtcnew,
                                   m_dt);
    }
    else
    {
      MayDay::Error("in AMRLevelGodunov::advance: m_coarser_level_ptr is not castable to AMRLevelGodunov*");
    }
  }
  // finer level exists but not coarser
  else if (m_finer_level_ptr != NULL)
  {
    AMRLevelGodunov* amrg_finer_ptr = dynamic_cast <AMRLevelGodunov*> (m_finer_level_ptr);
    if (amrg_finer_ptr != NULL)
    {
      const LevelData<FArrayBox> ld;
      LevelFluxRegister lfr;

      new_dt = m_levelgodunov.step(m_state_new,
                                   m_flux_register,
                                   lfr,
                                   ld,
                                   ld,
                                   m_time,
                                   Real(0.0),
                                   Real(0.0),
                                   m_dt
                                   );
    }
    else
    {
      MayDay::Error("in AMRLevelGodunov::advance: m_finer_level_ptr is not castable to AMRLevelGodunov*");
    }
  }
  // neither coarser nor finer levels exist
  else
  {
    const LevelData<FArrayBox> ld;
    LevelFluxRegister lfr;

    new_dt = m_levelgodunov.step(m_state_new,
                                 lfr,
                                 lfr,
                                 ld,
                                 ld,
                                 m_time,
                                 Real(0.0),
                                 Real(0.0),
                                 m_dt
                                 );
  }

  m_time += m_dt;
  Real return_dt = m_cfl * new_dt;

  m_dtNew = return_dt;

  return return_dt;
}

// things to do after a timestep
void AMRLevelGodunov::postTimeStep()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::postTimeStep " << m_level << endl;
  }

  // reflux
  if (m_flux_register.isDefined())
  {
    m_flux_register.reflux(m_state_new,1.0/m_dx);
  }

  // average from finer level data
  if (m_finer_level_ptr != NULL)
  {
    AMRLevelGodunov* amrg_ptr = dynamic_cast<AMRLevelGodunov*> (m_finer_level_ptr);
    if (amrg_ptr != NULL)
    {
      amrg_ptr->m_coarse_average.averageToCoarse(m_state_new,
                                                 amrg_ptr->m_state_new);
      Real mass = calculateMass();
      if (s_verbosity >= 3)
      {
        pout() << "level = " << m_level << "   mass = " << mass << endl;
      }
    }
    else
    {
      MayDay::Error("in AMRLevelGodunov::postTimeStep: m_coarser_level_ptr is not castable to AMRLevelGodunov*");
    }
  }

  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::postTimeStep " << m_level << " finished" << endl;
  }
}

// create tags for regridding
void AMRLevelGodunov::tagCells(IntVectSet& a_tags) 
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::tagCells " << m_level << endl;
  }

  // since tags are calculated using only current time step data, use
  // the same tagging function for initialization and for regridding.
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::tagCellsInit " << m_level << endl;
  }

  IntVectSet local_tags;
  // create tags based on undivided gradient of density
  const DisjointBoxLayout& level_domain = m_state_new.disjointBoxLayout();
  //this temporary data is no longer necessary
  LevelData<FArrayBox> state_with_ghost(level_domain,
                                        1,
                                        IntVect::Unit);

  // copy density
  m_state_new.copyTo(Interval(0,0),
                     state_with_ghost,
                     Interval(0,0));
  // fill ghost cells
  state_with_ghost.exchange(Interval(0,0));
  if (m_coarser_level_ptr != NULL)
  {
    const AMRLevelGodunov* amrg_ptr = dynamic_cast<AMRLevelGodunov*> (m_coarser_level_ptr);
    if (amrg_ptr != NULL)
    {
      PiecewiseLinearFillPatch pwl(level_domain,
                                   amrg_ptr->m_state_new.disjointBoxLayout(),
                                   m_num_comps,
                                   amrg_ptr->m_problem_domain,
                                   amrg_ptr->m_ref_ratio,
                                   1);

      pwl.fillInterp(state_with_ghost,
                     amrg_ptr->m_state_new,
                     amrg_ptr->m_state_new,
                     1.0,
                     0,
                     0,
                     1);
    }
    else
    {
      MayDay::Error("in AMRLevelGodunov::tagCellsInit: m_coarser_level_ptr is not castable to AMRLevelGodunov*");
    }
  }

  // compute undivided gradient
  DataIterator dit = level_domain.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
  {
    const Box& b = level_domain[dit()];
    FArrayBox grad_fab(b,SpaceDim);
    const FArrayBox& state_fab = state_with_ghost[dit()];
    for (int dir = 0; dir < SpaceDim; ++dir)
    {
      const Box bcenter = b & grow(m_problem_domain,-BASISV(dir));
      const Box blo = b & adjCellLo(bcenter,dir);
      const Box bhi = b & adjCellHi(bcenter,dir);
      const int haslo = ! blo.isEmpty();
      const int hashi = ! bhi.isEmpty();
      FORT_GETGRAD(CHF_BOX(bcenter),
                   CHF_BOX(blo),
                   CHF_BOX(bhi),
                   CHF_CONST_INT(dir),
                   CHF_CONST_INT(haslo),
                   CHF_CONST_INT(hashi),
                   CHF_CONST_FRA1(state_fab,0),
                   CHF_FRA1(grad_fab,dir));
    }
    FArrayBox grad_mag_fab(b,1);
    FORT_MAGNITUDE(CHF_BOX(b),
                   CHF_FRA1(grad_mag_fab,0),
                   CHF_CONST_FRA(grad_fab));

    // tag where gradient exceeds threshold
    BoxIterator bit(b);
    for (bit.begin(); bit.ok(); ++bit)
    {
      const IntVect& iv = bit();

      if (grad_mag_fab(iv) >= m_refine_thresh)
      {
        local_tags |= iv;
      }
    }
  }
  local_tags.grow(m_tag_buffer_size);
  // need to do this in two steps unless a IntVectSet::operator&=(ProblemDomain) operator is defined
  Box localTagsBox = local_tags.minBox();
  localTagsBox &= m_problem_domain;
  local_tags &= localTagsBox;

  Vector<IntVectSet> all_tags;

  const int dest_proc = uniqueProc(SerialTask::compute);

  gather(all_tags,local_tags,dest_proc);

  if (procID() == uniqueProc(SerialTask::compute))
  {
    for (int i = 0; i < all_tags.size(); ++i)
    {
      a_tags |= all_tags[i];
    }
  }
}

// create tags at initialization
void AMRLevelGodunov::tagCellsInit(IntVectSet& a_tags) 
{
  tagCells(a_tags);
}

// set up data on this level after regridding
void AMRLevelGodunov::regrid(const Vector<Box>& a_new_grids)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::regrid " << m_level << endl;
  }

  m_level_grids = a_new_grids;

  const DisjointBoxLayout level_domain = loadBalance(a_new_grids);

  if (s_verbosity >= 4)
  {
    pout() << "new grids: " << endl;
    for (LayoutIterator lit = level_domain.layoutIterator(); lit.ok(); ++lit)
    {
      pout() << level_domain[lit()] << endl;
    }
  }

  // save data for later copy
  LevelData<FArrayBox> old_state;
  old_state.define(m_state_new);
  m_state_new.copyTo(m_state_new.interval(),
                     old_state,
                     old_state.interval());

  // reshape state with new grids
  m_state_new.define(level_domain,m_num_comps,IntVect::Unit);
  m_state_old.define(level_domain,m_num_comps,IntVect::Unit);

  // set up data structures
  levelSetup(level_domain);

  // interpolate from coarser level
  if (m_coarser_level_ptr != NULL)
  {
    AMRLevelGodunov* amrg_ptr = dynamic_cast<AMRLevelGodunov*> (m_coarser_level_ptr);
    if (amrg_ptr != NULL)
    {
      m_fine_interp.interpToFine(m_state_new,
                                 amrg_ptr->m_state_new);
    }
    else
    {
      MayDay::Error("in AMRLevelGodunov::regrid: m_coarser_level_ptr is not castable to AMRLevelGodunov*");
    }
  }

  // copy from old state
  old_state.copyTo(old_state.interval(),
                   m_state_new,
                   m_state_new.interval());
}

// initialize grid
void AMRLevelGodunov::initialGrid(const Vector<Box>& a_new_grids)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::initialGrid " << m_level << endl;
  }

  m_level_grids = a_new_grids;

  const DisjointBoxLayout level_domain = loadBalance(a_new_grids);

  if (s_verbosity >= 4)
  {
    pout() << "new grids: " << endl;
    for (LayoutIterator lit = level_domain.layoutIterator(); lit.ok(); ++lit)
    {
      pout() << level_domain[lit()] << endl;
    }
  }
  m_state_new.define(level_domain,m_num_comps);
  m_state_old.define(level_domain,m_num_comps);

  // set up data structures
  levelSetup(level_domain);
}

// initialize data
void AMRLevelGodunov::initialData()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::initialData " << m_level << endl;
  }

  m_phys_ibc_ptr->initialize(m_state_new,
                             m_problem_domain,
                             m_dx);
}

// things to do after initialization
void AMRLevelGodunov::postInitialize()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::postInitialize " << m_level << endl;
  }
}

// set physical boundary conditions object
void AMRLevelGodunov::physIBC(RefCountedPtr<PhysIBC> a_phys_ibc_ptr)
{
  m_phys_ibc_ptr = a_phys_ibc_ptr;
}

#ifdef HDF5

// write checkpoint header
void AMRLevelGodunov::writeCheckpointHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::writeCheckpointHeader" << endl;
  }

  HDF5HeaderData header;
  header.m_int["num_components"] = m_num_comps;

  char comp_str[30];
  for (int comp = 0; comp < m_num_comps; ++comp)
  {
    sprintf(comp_str,"component_%d",comp);
    header.m_string[comp_str] = m_state_names[comp];
  }
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }
}

// write level to checkpoint file
void AMRLevelGodunov::writeCheckpointLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::writeCheckpointLevel" << endl;
  }

  char level_str[20];
  sprintf(level_str,"%d",m_level);
  const std::string label = std::string("level_") + level_str;

  a_handle.setGroup(label);

  HDF5HeaderData header;

  header.m_int ["ref_ratio"]       = m_ref_ratio;
  header.m_int ["tag_buffer_size"] = m_tag_buffer_size;
  header.m_real["dx"]              = m_dx;
  header.m_real["dt"]              = m_dt;
  header.m_box ["prob_domain"]     = m_problem_domain.domainBox();

  // write out periodicity info
  
  D_TERM(
         if (m_problem_domain.isPeriodic(0))
           header.m_int ["is_periodic_0"] = 1;
         else
           header.m_int ["is_periodic_0"] = 0; , 
    
         if (m_problem_domain.isPeriodic(1))
           header.m_int ["is_periodic_1"] = 1;
         else
           header.m_int ["is_periodic_1"] = 0; , 
         
         if (m_problem_domain.isPeriodic(2))
           header.m_int ["is_periodic_2"] = 1;
         else 
           header.m_int ["is_periodic_2"] = 0; );

      
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }

  write(a_handle,m_state_new.boxLayout());
  write(a_handle,m_state_new,"data");
}

// read header from checkpoint file
void AMRLevelGodunov::readCheckpointHeader(HDF5Handle& a_handle)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::readCheckpointHeader" << endl;
  }

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << "hdf5 header data:" << endl;
    pout() << header << endl;
  }

  // read number of components
  if (header.m_int.find("num_components") == header.m_int.end())
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointHeader: checkpoint file does not have num_components");
  }

  int num_comps = header.m_int["num_components"];
  if (num_comps != m_num_comps)
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointHeader: num_components in checkpoint file does not match solver");
  }

  // read component names
  std::string state_name;
  char comp_str[60];
  for (int comp = 0; comp < m_num_comps; ++comp)
  {
    sprintf(comp_str,"component_%d",comp);
    if (header.m_string.find(comp_str) == header.m_string.end())
    {
      MayDay::Error("AMRLevelGodunov::readCheckpointHeader: checkpoint file does not have enough component names");
    }

    state_name = header.m_string[comp_str];
    if (state_name != m_state_names[comp])
    {
      MayDay::Error("AMRLevelGodunov::readCheckpointHeader: state_name in checkpoint does not match solver");
    }
  }
}

// read level data from checkpoint file
void AMRLevelGodunov::readCheckpointLevel(HDF5Handle& a_handle)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::readCheckpointLevel" << endl;
  }

  // char* level_str = new char[(int)(log10(m_level+1))+1];
  char level_str[20];
  sprintf(level_str,"%d",m_level);
  const std::string label = std::string("level_") + level_str;
  // delete[] level_str;

  a_handle.setGroup(label);

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << "hdf5 header data:" << endl;
    pout() << header << endl;
  }

  // read refinement ratio
  if (header.m_int.find("ref_ratio") == header.m_int.end())
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointLevel: file does not contain ref_ratio");
  }
  m_ref_ratio = header.m_int["ref_ratio"];

  if (s_verbosity >= 2)
  {
    pout() << "read ref_ratio = " << m_ref_ratio << endl;
  }

  if (header.m_int.find("tag_buffer_size") == header.m_int.end())
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointLevel: file does not contain tag_buffer_size");
  }
  m_tag_buffer_size=  header.m_int["tag_buffer_size"];

  if (s_verbosity >= 2)
  {
    pout() << "read tag_buffer_size = " << m_tag_buffer_size << endl;
  }

  // read dx
  if (header.m_real.find("dx") == header.m_real.end())
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointLevel: file does not contain dx");
  }
  m_dx = header.m_real["dx"];

  if (s_verbosity >= 2)
  {
    pout() << "read dx = " << m_dx << endl;
  }

  // read dt
  if (header.m_real.find("dt") == header.m_real.end())
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointLevel: file does not contain dt");
  }
  m_dt = header.m_real["dt"];

  if (s_verbosity >= 2)
  {
    pout() << "read dt = " << m_dt << endl;
  }

  // read problem domain
  if (header.m_box.find("prob_domain") == header.m_box.end())
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointLevel: file does not contain prob_domain");
  }
  Box domainBox = header.m_box["prob_domain"];
      

  // read in periodicity info -- this is more complicated than
  // it really needs to be in order to preserve backward compatibility 
  bool is_periodic[SpaceDim];
  D_TERM(
         if (!(header.m_int.find("is_periodic_0") == header.m_int.end()))
           is_periodic[0] =  (header.m_int["is_periodic_0"] == 1);
         else
           is_periodic[0] = false; , 

         if (!(header.m_int.find("is_periodic_1") == header.m_int.end()))
           is_periodic[1] =  (header.m_int["is_periodic_1"] == 1);
         else
           is_periodic[1] = false; , 

         if (!(header.m_int.find("is_periodic_2") == header.m_int.end()))
           is_periodic[2] =  (header.m_int["is_periodic_2"] == 1);
         else
           is_periodic[2] = false; );

  m_problem_domain = ProblemDomain(domainBox, is_periodic);

  // read grids
  Vector<Box> grids;
  const int grid_status = read(a_handle,grids);
  if (grid_status != 0)
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointLevel: file does not contain a Vector<Box>");
  }

  // create level domain
  const DisjointBoxLayout level_domain = loadBalance(grids);

  LayoutIterator lit = level_domain.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit)
  {
    const Box& b = level_domain[lit()];
    m_level_grids.push_back(b);
  }

  if (s_verbosity >= 4)
  {
    pout() << "read level domain: " << endl;
    LayoutIterator lit = level_domain.layoutIterator();
    for (lit.begin(); lit.ok(); ++lit)
    {
      const Box& b = level_domain[lit()];
      pout() << lit().intCode() << ": " << b << endl;
    }
    pout() << endl;
  }

  // reshape state with new grids
  m_state_new.define(level_domain,m_num_comps);
  const int data_status = read<FArrayBox>(a_handle,
                                          m_state_new,
                                          "data",
                                          level_domain);
  if (data_status != 0)
  {
    MayDay::Error("AMRLevelGodunov::readCheckpointLevel: file does not contain state data");
  }
  m_state_old.define(level_domain,m_num_comps);

  // set up data structures
  levelSetup(level_domain);
}

// write header data to plotfile
void AMRLevelGodunov::writePlotHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::writePlotHeader" << endl;
  }

  HDF5HeaderData header;
  header.m_int["num_components"] = m_num_comps;

  char comp_str[30];
  for (int comp = 0; comp < m_num_comps; ++comp)
  {
    sprintf(comp_str,"component_%d",comp);
    header.m_string[comp_str] = m_state_names[comp];
  }
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }
}

// write level data to plotfile
void AMRLevelGodunov::writePlotLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::writePlotLevel" << endl;
  }

  // char* level_str = new char[(int)(log10(m_level+1))+1];
  char level_str[20];
  sprintf(level_str,"%d",m_level);
  const std::string label = std::string("level_") + level_str;
  // delete[] level_str;

  a_handle.setGroup(label);

  HDF5HeaderData header;

  header.m_int ["ref_ratio"]   = m_ref_ratio;
  header.m_real["dx"]          = m_dx;
  header.m_real["dt"]          = m_dt;
  header.m_box ["prob_domain"] = m_problem_domain.domainBox();
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }

  //if (m_coarser_level_ptr != NULL)
  //  {
  //    const AMRLevelGodunov* coarse = (const  AMRLevelGodunov*)m_coarser_level_ptr;
  //    m_fine_interp.interpToFine(m_state_new,coarse->m_state_new);
  //  }

  write(a_handle,m_state_new.boxLayout());
  write(a_handle,m_state_new,"data");
}

#endif

// compute maximum stable dt for this level (does not actually _set_ m_dt)
Real AMRLevelGodunov::computeDt()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::computeDt " << m_level << endl;
  }

  Real newDt;
  newDt = m_dtNew;

  return newDt;
}

// compute dt with initial data
Real AMRLevelGodunov::computeInitialDt()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::computeInitialDt " << m_level << endl;
  }

  Real newDT = m_initial_dt_multiplier * m_dx / m_levelgodunov.getMaxWaveSpeed(m_state_new);

  return newDT;
}

// set the CFL number
void AMRLevelGodunov::CFL(Real a_cfl)
{
  m_cfl = a_cfl;
}

// set the refinement threshold
void AMRLevelGodunov::refinementThreshold(Real a_refine_thresh)
{
  m_refine_thresh = a_refine_thresh;
}

// set the tag buffer size
void AMRLevelGodunov::tagBufferSize(int a_tag_buffer_size)
{
  m_tag_buffer_size = a_tag_buffer_size;
}

// create a load-balanced DisjointBoxLayout from a collection of Boxes
DisjointBoxLayout AMRLevelGodunov::loadBalance(const Vector<Box>& a_grids)
{
  // load balance and create boxlayout
  Vector<int> proc_map;
  if (procID() == uniqueProc(SerialTask::compute))
  {
    LoadBalance(proc_map,a_grids);
  }
  broadcast(proc_map,uniqueProc(SerialTask::compute));

  if (s_verbosity >= 4)
  {
    pout() << "AMRLevelGodunov::loadBalance: procesor map: " << endl;
    for (int igrid = 0; igrid < a_grids.size(); ++igrid)
    {
      pout() << igrid << ": " << proc_map[igrid] << "  " << endl;
    }
    pout() << endl;
  }

  DisjointBoxLayout dbl(a_grids,proc_map, m_problem_domain);
  dbl.close();

  return dbl;
}

// calculate "mass" by summing density
Real AMRLevelGodunov::calculateMass() const
{
  Real local_mass = 0;
  DataIterator dit = m_state_new.boxLayout().dataIterator();

  for (dit.begin(); dit.ok(); ++dit)
  {
    local_mass += m_state_new[dit()].sum(m_state_new.box(dit()),0,1);
  }

  Vector<Real> all_masses;
  gather(all_masses,local_mass,uniqueProc(SerialTask::compute));
  Real global_mass = 0.0;

  if (procID() == uniqueProc(SerialTask::compute))
  {
    for (int i = 0; i < all_masses.size(); ++i)
    {
      global_mass += all_masses[i];
    }
  }

  broadcast(global_mass,uniqueProc(SerialTask::compute));
  for (int d = 0; d < SpaceDim; ++d)
  {
    global_mass *= m_dx;
  }

  return global_mass;
}

// set up menagerie of data structures
void AMRLevelGodunov::levelSetup(const DisjointBoxLayout& a_level_domain)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelGodunov::levelSetup " << m_level << endl;
  }

  if (m_coarser_level_ptr != NULL)
  {
    int nRefCrse = m_coarser_level_ptr->refRatio();

    m_coarse_average.define(a_level_domain,
                            m_num_comps,
                            nRefCrse);
    m_fine_interp.define(a_level_domain,
                         m_num_comps,
                         nRefCrse,
                         m_problem_domain);

    AMRLevelGodunov* amrg_ptr = dynamic_cast <AMRLevelGodunov*> (m_coarser_level_ptr);
    if (amrg_ptr != NULL)
    {
      const DisjointBoxLayout& coarser_level_domain = amrg_ptr->m_state_new.disjointBoxLayout();

      // maintain levelgodunov
      m_levelgodunov.define(a_level_domain,
                            coarser_level_domain,
                            m_problem_domain,
                            4,
                            nRefCrse,
                            m_dx,
                            m_phys_ibc_ptr);
      // maintain flux registers
      amrg_ptr->m_flux_register.undefine();
      amrg_ptr->m_flux_register.define(a_level_domain,
                                       coarser_level_domain,
                                       m_problem_domain,
                                       amrg_ptr->m_ref_ratio,
                                       m_num_comps);
      amrg_ptr->m_flux_register.setToZero();
    }
    else
    {
      MayDay::Error("in AMRLevelGodunov::levelSetup: m_coarser_level_ptr is not castable to AMRLevelGodunov*");
    }
  }
  else
  {
    m_levelgodunov.define(a_level_domain,
                          DisjointBoxLayout(),
                          m_problem_domain,
                          4,
                          m_ref_ratio,
                          m_dx,
                          m_phys_ibc_ptr);
  }

  if (m_finer_level_ptr == NULL)
  {
    m_flux_register.undefine();
  }
}
