/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//  PhysIBC.H
//============
//
// Pure virtual base class through which a user
// specifies the initial and boundary conditions for LevelGodunov
//
//  For boundary conditions:
//
//  For every box in this level, the virtual function
//
//  fluxBC(conState, primState, artVisc, qGdnv, dir,
//                  box, time, dt, dx, DProblem, flux)
//
//  is called.
//
//  conState:  The conserved variables at the
//             previous time step.  Includes ghost cells
//             defined at (x,time).
//  primState: The primitive variables at the
//             previous time step.  Includes ghost cells
//             defined at (x,time).
//  artVisc:   Artificial viscosity at (x+-dx, time).
//  qGdnv:     Godunov-scheme derived values for edge based
//             primitive variables in the current drection.
//             qGdnv is the predicted values of primState
//             at (x+-dx/2, time+-dt/2)
//  dir:       coordinate direction that we currently need BC for.
//  box:       Box over which U is actually defined at this level
//  time:      the previous time value at which conState and PrimState
//             are defined.
//  dt:        current timestep
//  dx:        mesh spacing at this level
//  DProblem:  Box that defines the domain boundary
//             in the index space of this level.
//  flux:      The current values of the conservative flux
//             that has been computed without boundary conditions.
//
// original design        : Phil Colella and Brian Van Straalen
// original implementation: Brian Van Straalen
// implementation date    : 06/25/99
// latest revision        : 12/30/99 (PC)
//

#ifndef _PHYSIBC_H
#define _PHYSIBC_H

#include "FArrayBox.H"
#include "REAL.H"
#include "LevelData.H"
#include "ProblemDomain.H"

///
/**
   Pure virtual base class through which a user specifies the
   initial and boundary conditions for LevelGodunov.
 */
class PhysIBC
{
public:
  ///
  PhysIBC() {};

  ///
  virtual ~PhysIBC() {};

  ///  For every box in this level, this virtual function is called.
  virtual void fluxBC(FArrayBox&       a_flux,
                      const FArrayBox& a_conState,
                      const FArrayBox&       a_artVisc,
                      const FArrayBox&       a_qGdnv,
                      const FArrayBox&       a_prim,
                      const int&        a_dir,
                      const Real&      a_time,
                      const Real&      a_dt,
                      const Real&      a_dx,
                      const Box&       a_DProblem
                      ) const = 0;

  ///  For every box in this level, this virtual function is called.
  virtual void fluxBC(FArrayBox&       a_flux,
                      const FArrayBox& a_conState,
                      const FArrayBox&       a_artVisc,
                      const FArrayBox&       a_qGdnv,
                      const FArrayBox&       a_prim,
                      const int&        a_dir,
                      const Real&      a_time,
                      const Real&      a_dt,
                      const Real&      a_dx,
                      const ProblemDomain&       a_DProblem) const = 0;

  /// Set boundary slopes
  virtual void setBdrySlopes(FArrayBox&       a_dq,
                             const Box&       a_DProblem,
                             const Real&      a_dx,
                             const int&       a_dir,
                             const FArrayBox& a_prim
                             ) const = 0;


  /// Set boundary slopes
  virtual void setBdrySlopes(FArrayBox&       a_dq,
                             const ProblemDomain&  a_DProblem,
                             const Real&      a_dx,
                             const int&       a_dir,
                             const FArrayBox& a_prim)const = 0;


  /// Initialize
  virtual void initialize(LevelData<FArrayBox>& a_conState,
                          const Box&                   a_DProblem,
                          const Real&                  a_dx) const = 0;


  /// Initialize
  virtual void initialize(LevelData<FArrayBox>& a_conState,
                          const ProblemDomain&        a_DProblem,
                          const Real&                  a_dx) const = 0;
};

#endif
