/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Tues, July 6, 1999

#ifndef _AMRLevelMG_H_
#define _AMRLevelMG_H_

#include <stdlib.h>
#include <iostream>
#include <assert.h>
#include <math.h>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "LevelFluxRegister.H"
#include "BaseFab.H"
#include "LevelMG.H"
#include "MGInterp.H"
#include "Copier.H"

//forward declaration of AMRSolver
class AMRSolver;

/// class which manages data and operations on a level for AMRSolver
/**
   AMRLevelMG manages the operations on a level for AMRSolver.
   AMRSolver contains a vector of AMRLevelMG's (one for each level).
   This class should be considered internal to AMRSolver and should
   not be considered part of the Chombo API.
 */
class AMRLevelMG
{
public:    
  friend class AMRSolver;

  /// 
  AMRLevelMG();

  /// 
  ~AMRLevelMG();

  /// 
  /**
     Define AMRLevelMG from the inputs.
  */
  AMRLevelMG(const AMRSolver* const a_parent, 
             int a_level, 
             const LevelOp* const a_opin,
             int ncomp=1);

  ///
  /**
     Define AMRLevelMG from the inputs.
     The AMRLevelMG will only work for the 
     input AMRSolver.  A LevelMG is created.
  */
  void define(const AMRSolver* const a_parent, 
              int a_level, 
              const LevelOp* const a_opin,
              int ncomp=1);

  ///
  /** Set number of multigrid smoothings on way up V-cycle;
      Default == 4.
  */
  void setnumSmoothUp(int a_numSmoothUp);

  ///
  /** Set number of multigrid smoothings on way down V-cycle;
      Default == 4.
  */
  void setnumSmoothDown(int a_numSmoothDown);

  /// 
  /**
     Complete, 3-level operator.  Applies operator with
     inhomogeneous coarse-fine boundary conditions (and
     inhomogeneous domain boundary conditions.  Refluxing
     is used to compute the operator at coarse-fine boundaries.
  */
  void applyAMROperator(Vector<LevelData<FArrayBox> *>& a_phiLevel, 
                        LevelData<FArrayBox> & Lofphi);


  /// 
  /** Complete, 3-level operator with homogeneous physical 
      boundary conditions.  Applies operator with inhomogeneous
      physical boundary conditions.  Refluxing is used to compute 
      the operator at coarse-fine boundaries.
  */
  void applyAMROperatorHphys(Vector<LevelData<FArrayBox> *>& a_phiLevel,
                             LevelData<FArrayBox>& LofPhi);

  ///
  /** 
      Compute complete, 3-level residual 
      and put it into local data m_resid
  */
  void computeAMRResidual(Vector<LevelData<FArrayBox> *>& a_phiLevel,
                          const Vector<LevelData<FArrayBox> *>& a_rhsLevel);


  ///
  /** 
      Compute complete, 3-level residual with homogeneous physical
      BCs and inhomogeneous C/F BCs and put it into local data m_resid
  */
  void computeAMRResidualHphys(Vector<LevelData<FArrayBox> *>& a_phiLevel,
                               const Vector<LevelData<FArrayBox> *>& a_rhsLevel);


  ///
  /**
     Compute complete, 3-level residual 
     and return it in the argument LevelData<FArrayBox>
  */
  void computeAMRResidual(LevelData<FArrayBox>& a_resid,
                          Vector<LevelData<FArrayBox>* >& a_phiLevel,
                          const Vector<LevelData<FArrayBox>* >& a_rhsLevel);
  

  ///
  /** 
      Sweep up multigrid v-cycle.  Computes residual
      on 
  */
  void upSweep(Vector<LevelData<FArrayBox> *>& a_phiLevel,
               const Vector<LevelData<FArrayBox> *>& a_rhsLevel);

  ///
  /** 
      Sweep down v-cycle
  */
  void downSweep(Vector<LevelData<FArrayBox> *>& a_phiLevel,
                 const Vector<LevelData<FArrayBox> *>& a_rhsLevel);

  ///
  /** 
      compute and return norm of internal data m_resid.
  */
  Vector<Real> computeResidualNorm(int normType) const;

  // 
  LevelOp* levelOpPtr() const;


  /// 
  bool isDefined() const;

protected:

  // returns normType norm of mfab 
  Vector<Real>
  computeNorm(const LevelData<FArrayBox> & a_mfinput, 
              int a_normType) const;

  /*  correct the input a_soln with a_crseCorr */
  void 
  crseCorrect(
              LevelData<FArrayBox> & a_soln,
              const LevelData<FArrayBox> & a_crseCorr,
              int a_nref) ;

  //use short v-cycle or levelmg's smoother to reduce
  //high wave numbers of residual
  void smooth(LevelData<FArrayBox> & a_soln, 
              const LevelData<FArrayBox> & rhs);

  //delete internal memory and set pointers to null
  void clearMemory();

  //set default values.  does not deal with memory
  void setDefaultValues();

  /*
    reflux enforces flux-matching from finer levels.
    Steps involved:
    a)initialize flux register with coarse flux
    b)increment flux register with sum of fine fluxes
    b)modify solution with flux diffrence
  */
  void
  reflux(Vector<LevelData<FArrayBox> *>& a_phiLevel,
         LevelData<FArrayBox> & a_Lofphi);

  /*
    Fill fluxregisters with coarse flux
  */
  void
  initFRCoarse(Vector<LevelData<FArrayBox> *>& a_phiLevel);

  /*
    increment fluxregisters with fine flux
  */
  void
  incrementFRFine(Vector<LevelData<FArrayBox> *>& a_phiLevel);

  // da boss
  const AMRSolver* m_parent;

  // LevelMG level solver object to relax on this level
  LevelMG m_levelMG;

  // for interpolation
  MGInterp m_mginterp;

  //flux register with next coarser level
  LevelFluxRegister m_levfluxreg;

  // has this AMRLevelMG been defined yet?
  bool m_isDefined;

  DisjointBoxLayout m_coarsenedGrids;
  // residual
  LevelData<FArrayBox> m_resid;

  // correction
  LevelData<FArrayBox> m_corr;

  // the infamous correction to the correction
  LevelData<FArrayBox> m_dcorr;

  // L(phi)
  LevelData<FArrayBox> m_lofPhi;

  // extra copy of solution
  LevelData<FArrayBox> m_phiSave;

  LevelData<FArrayBox> m_resC;

  // Copier that operates for moving  coarsened residual to 
  // coarse level grid space.
  Copier m_residualCopier;

  // Copier that's used for exchange on the fine grid data
  // before fine flux register data is incremented.
  Copier m_fineExchangeCopier;

  //
  int m_level;

  //
  LevelOp* m_levelopPtr;

  //
  CoarseAverage m_averageOp;

  bool m_arrayViewVerbose;
private:
  void  setArrayViewVerbose(bool a_verbosity);
  AMRLevelMG& operator = (const AMRLevelMG&);
  AMRLevelMG(const AMRLevelMG&);

};

#endif
