/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Mon, July 19, 1999

#ifndef _GHOST_BC_H_
#define _GHOST_BC_H_

#include "Box.H"
#include "FArrayBox.H"
#include "REAL.H"
#include "SPACE.H"
#include "Tuple.H"

#include "Interval.H"
#include "Vector.H"
#include "LoHiSide.H"
#include "ProblemDomain.H"

///class to encapsulate the operations  of ghost-cell boundary conditions.

/**
BoxGhostBC is a class to encapsulate the 
operations  of ghost-cell boundary conditions at a face.  
If the solution is phi and the face normal direction is x,
the boundary conditions usually used can be expressed as
A*phi + B*dphi/dx = C.
The derived class provides a function to 
to fill A B and C.
These functions are only meant to be called by DomainGhostBC.
*/
class BoxGhostBC
{
public:
  friend class DomainGhostBC;

  ///
  virtual ~BoxGhostBC(){;}

  ///
  BoxGhostBC():m_components(-1,-1){};

  /// sets interval to be (0,0) 
  BoxGhostBC(int a_dir, Side::LoHiSide a_sd);


  ///
  BoxGhostBC(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps);


protected:

  /// virtual constructor workaround.
  virtual BoxGhostBC* new_boxghostbc() const = 0;

  /// sets interval to be (0,0)
  void
  define(int a_dir, Side::LoHiSide a_sd);


  ///
  void
  define(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps);

  ///
  virtual void
  applyInhomogeneousBCs(FArrayBox& a_state, 
                        const Box& a_domain, 
                        Real a_dx) const; 

  ///
  virtual void
  applyHomogeneousBCs( FArrayBox& a_state, 
                       const Box& domain, 
                       Real a_dx) const;


  ///
  virtual void
  applyInhomogeneousBCs(FArrayBox& a_state, 
                        const ProblemDomain& a_domain, 
                        Real a_dx) const; 

  ///
  virtual void
  applyHomogeneousBCs( FArrayBox& a_state, 
                       const ProblemDomain& a_domain, 
                       Real a_dx) const;

  ///
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
               FArrayBox& a_dircfac,
               FArrayBox& a_inhmval,
               Real a_dx,
               const Box& domain) const = 0;

  virtual void
  applyBCs(const Box& a_bcbox, 
           FArrayBox& a_state,
           const FArrayBox& a_neumfac,
           const FArrayBox& a_dircfac,
           const FArrayBox& a_inhmval,
           Real a_dx) const;

  ///
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
               FArrayBox& a_dircfac,
               FArrayBox& a_inhmval,
               Real a_dx,
               const ProblemDomain& domain) const = 0;


  Side::LoHiSide m_side;
  int m_direction;
  Interval m_components;

private:
  BoxGhostBC(const BoxGhostBC&):m_components(-1,-1) {;}
  virtual void operator=(const BoxGhostBC&){;}
};


/// Class to enforce ghost-cell based boundary conditions
/**
   This class which holds a domain's worth of boxghostbcs,
   one for each face.  The user adds a BoxGhostBC-derived
   for each face to enforce boundary conditions.
   If the solution is phi and the face normal direction is x,
   the boundary conditions at a face are  expressed as
   A*phi + B*dphi/dx = C.
   The class derived from BoxGhostBC provides a function to 
   to fill A B and C.  DomainGhostBC calls this function
   and fills the ghost cells outside the domain appropriately.
 */
class DomainGhostBC
{
public:
  ///
  DomainGhostBC();

  ///
  ~DomainGhostBC();

  ///
  /**
     Set boundary conditions at a face 
     with ghost_bc.  
   */
  void
  setBoxGhostBC(const BoxGhostBC& ghost_bc);

  ///
  const BoxGhostBC&
  operator() (int direction, Side::LoHiSide side) const;

  ///
  /**
     apply boundary conditions on all sides
     compares sides of state to domain.
  */
  void
  applyHomogeneousBCs(FArrayBox& state,
                      const Box& Domain, 
                      Real dx) const;

  ///
  /**
     apply boundary conditions on all sides
     compares sides of state to domain.
  */
  void
  applyHomogeneousBCs(FArrayBox& state,
                      const ProblemDomain& Domain, 
                      Real dx) const;

  ///
  /**
     apply boundary conditions on all sides
     compares sides of state to domain.
  */
  void
  applyInhomogeneousBCs(FArrayBox& state,
                        const Box& Domain, 
                        Real dx) const;


  ///
  /**
     apply boundary conditions on all sides
     compares sides of state to domain.
  */
  void
  applyInhomogeneousBCs(FArrayBox& state,
                        const ProblemDomain& Domain, 
                        Real dx) const;

  ///
  DomainGhostBC& operator=(const DomainGhostBC&);

  ///
  DomainGhostBC(const DomainGhostBC&);
protected:

  /// returns true if this GhostBC is defined
  bool isBCDefined(const int a_dir, const Side::LoHiSide a_side) const;
  void resetBoxGhostBC(const int a_dir, const Side::LoHiSide a_side);

  Tuple<BoxGhostBC*, SpaceDim> m_loGhostBC;
  Tuple<BoxGhostBC*, SpaceDim> m_hiGhostBC;
private:
};

#endif
