/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Mon, July 26, 1999

#ifndef _LevelOP_H_
#define _LevelOP_H_

#include <iostream>
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include <REAL.H>
#include <IntVect.H>
#include <Box.H>
#include <DisjointBoxLayout.H>
#include <LevelData.H>
#include <FArrayBox.H>
#include <LoHiSide.H>
#include <ProblemDomain.H>

/// Pure base class to encapsulate level operations API for elliptic solvers.
/**
   LevelOp is a pure base class to 
   encapsulate level operations API for elliptic solvers.
   All elliptic operators used by AMRSolver inherit the LevelOp interface.
*/
class LevelOp
{

public:

  ///constructors, destructors, defines

  /// 
  /**
     Coarse / Fine interpolation operator.
     Fill ghost cells of phi
  */
  virtual
  void CFInterp(
                LevelData<FArrayBox>& a_phi,
                const LevelData<FArrayBox>& a_phiCoarse
                ) = 0;

  /// 
  /**
     homogeneous coarse/fine interpolation operator
  */
  virtual
  void homogeneousCFInterp(LevelData<FArrayBox>& a_phi) = 0;

  ///
  LevelOp() {};

  ///
  /**
     this gets around the "no virtual constructor" rule
  */
  virtual LevelOp* new_levelop() const = 0;

  ///
  virtual bool isDefined() const = 0;

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  virtual void define(
                      const DisjointBoxLayout& Ba,
                      const DisjointBoxLayout* base_ba,
                      Real  DxLevel, 
                      int refratio,
                      const Box& domf,
                      bool homogeneousOnly,
                      int ncomp = 1) = 0;

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  virtual void define(
                      const DisjointBoxLayout& Ba,
                      const DisjointBoxLayout* base_ba,
                      Real  DxLevel, 
                      int refratio,
                      const ProblemDomain& domf,
                      bool homogeneousOnly,
                      int ncomp = 1) = 0;


  ///
  /**
     Full define function.  define
     from finer levelop.
     reftofine is the refinement ratio
     to get from *this opfine
  */
  virtual void define(
                      const LevelOp* opfine,
                      int reftofine) = 0;

  ///
  virtual ~LevelOp(){};

  ///
  /**
     Smoother. 
     Assumes that problem has already been put in 
     residual correction form, 
     so that C/F boundary conditions are homogeneous.
  */
  virtual void smooth(
                      LevelData<FArrayBox>& phi,
                      const LevelData<FArrayBox>& rhs
                      ) = 0;

  
  ///
  /** apply preconditioning to the solution on this level.
      Given rhshat, returns phihat from M(phihat) = rhshat
  */
  virtual void levelPreconditioner(
                                   LevelData<FArrayBox>& a_phihat,
                                   const LevelData<FArrayBox>& a_rhshat
                                   ) = 0;


  ///
  /**  
       Evaluate Operator, inhomogeneous C/F boundary conditions.
       inhomogeneous phys boundary conditions.
  **/
  virtual void applyOpI(
                        LevelData<FArrayBox>& phi,
                        const LevelData<FArrayBox>* phiCoarse,
                        LevelData<FArrayBox>& LOfPhi
                        ) = 0;

  ///
  /**  
       Evaluate Operator, inhomogeneous C/F boundary conditions.
       homogeneous phys boundary conditions.
  **/
  virtual void applyOpIcfHphys(
                               LevelData<FArrayBox>& phi,
                               const LevelData<FArrayBox>* phiCoarse,
                               LevelData<FArrayBox>& LOfPhi
                               )  = 0;

  ///
  /**  
       Evaluate Operator, homogeneous C/F boundary conditions.
       inhomogeneous phys boundary conditions.
  */
  virtual void applyOpHcfIphys(
                               LevelData<FArrayBox>& phi,
                               LevelData<FArrayBox>& LOfPhi
                               ) = 0;

  ///
  /**  
       Evaluate Operator, homogeneous C/F boundary conditions.
       homogeneous phys boundary conditions.
  */
  virtual void applyOpH(
                        LevelData<FArrayBox>& phi,
                        LevelData<FArrayBox>& LOfPhi
                        ) = 0;

  ///
  /**
     smoother at bottom level 
  */
  virtual void bottomSmoother(
                              LevelData<FArrayBox>& phi,
                              const LevelData<FArrayBox>& rhs) = 0;

  ///
  /** 
      get flux( == flux at THIS level) 
      The fluxes live on the cell faces with direction dir.
      Fluxes are computed for all interior edges of data.
      The flux fab is resized inside the routine.
  */
  virtual void 
  getFlux(
          FArrayBox& flux, 
          const FArrayBox& data, 
          const DataIndex& a_datInd,
          int dir) = 0;

};




#endif

