/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Weds, July 21, 1999

#ifndef _POISSONOP_H_
#define _POISSONOP_H_

#include <iostream>
#include <cmath>
#include <assert.h>
#include <cstdlib>
#include <REAL.H>
#include <IntVect.H>
#include <Box.H>
#include <DisjointBoxLayout.H>
#include <LevelData.H>
#include <BaseFab.H>
#include <FArrayBox.H>
#include "GhostBC.H"
#include "LevelOp.H"
#include "QuadCFInterp.H"
#include "CFIVS.H"
#include "Copier.H"
#include "BaseBottomSmoother.H"

///  one-sided differencing version of poisson operator
/** this class contains stuff to solve  div (grad phi) = rhs
 */
class PoissonOp: public LevelOp
{
public:

  ///
  /**
     virtual constructor workaround
  */
  virtual LevelOp* new_levelop() const;

  ///
  virtual bool isDefined() const;

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  void define(
              const DisjointBoxLayout& Ba,
              const DisjointBoxLayout* base_ba,
              Real  DxLevel, 
              int refratio,
              const Box& domf,
              bool a_homogeneousOnly=false,
              int ncomp = 1);

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  void define(
              const DisjointBoxLayout& Ba,
              const DisjointBoxLayout* base_ba,
              Real  DxLevel, 
              int refratio,
              const ProblemDomain& domf,
              bool a_homogeneousOnly=false,
              int ncomp = 1);

  ///
  /**
     Full define function.  define
     from finer levelop.
  */
  void define(
              const LevelOp* opfine,
              int refratio);

  ///
  PoissonOp();

  ///
  ~PoissonOp();

  ///
  /**
     Smoother. 
     Assumes that problem has already been put in 
     residual correction form, 
     so that C/F boundary conditions are homogeneous.
  */
  void smooth(
              LevelData<FArrayBox>& a_phi,
              const LevelData<FArrayBox>& a_rhs);


  ///
  /** apply preconditioning to the solution on this level.
      Given rhshat, returns phihat from M(phihat) = rhshat
  */
  virtual void levelPreconditioner(
                                   LevelData<FArrayBox>& a_phihat,
                                   const LevelData<FArrayBox>& a_rhshat
                                   );


  ///
  /**
     Evaluate Operator, inhomogeneous C/F boundary conditions.
     and inhomogeneous phys boundary conditions
  */
  void applyOpI(
                LevelData<FArrayBox>& a_phi,
                const LevelData<FArrayBox>* a_phiCoarse,
                LevelData<FArrayBox>& a_LOfPhi);

  /// 
  /**
     Evaluate Operator, inhomogeneous C/F boundary conditions.
     with homogeneous physical boundary conditions
  */
  void applyOpIcfHphys(
                       LevelData<FArrayBox>& a_phi,
                       const LevelData<FArrayBox>* a_phiCoarse,
                       LevelData<FArrayBox>& a_LOfPhi);

  ///
  /**
     Evaluate Operator, homogeneous C/F boundary conditions.
     and homogeneous physical boundary conditions
  */
  void applyOpH(
                LevelData<FArrayBox>& a_phi,
                LevelData<FArrayBox>& a_LOfPhi);

  /// 
  /**
     Coarse / Fine interpolation operator.
  */
  void CFInterp(LevelData<FArrayBox>& a_phi,
                const LevelData<FArrayBox>& a_phiCoarse
                );

  ///
  /** 
      Evaluate Operator, homogeneous C/F boundary conditions.
      with inhomogeneous physical boundary conditions
  */
  void applyOpHcfIphys(
                       LevelData<FArrayBox>& a_phi,
                       LevelData<FArrayBox>& a_lOfPhi);


  /// 
  /**
     homogeneous coarse/fine interpolation operator
  */
  void homogeneousCFInterp(LevelData<FArrayBox>& a_phi);

  ///
  void setDomainGhostBC(const DomainGhostBC& a_dombcin);

  /// 
  void setBottomSmoother(const BaseBottomSmoother& a_bottomSmoother);

  ///
  /**
     does homogeneous coarse/fine interpolation
  */
  void
  homogeneousCFInterp(LevelData<FArrayBox>& a_phif, 
                      const DataIndex& a_datInd, 
                      int a_idir, 
                      Side::LoHiSide a_hiorlo);

  ///
  /**
     This does a GSRB Pre/Conditioned BiCGStab on a level
     for the bottom solver.  
  */
  void bottomSmoother(
                      LevelData<FArrayBox>& a_phi,
                      const LevelData<FArrayBox>& a_rhs);

  ///
  /** 
      get flux( == flux at THIS level) 
      The fluxes live on the cell faces with direction dir.
      Fluxes are computed for all interior edges of data.
      The flux fab is resized inside the routine.
  */
  void getFlux(
               FArrayBox& flux, 
               const FArrayBox& a_data, 
               const DataIndex& a_datInd,
               int a_dir);

protected:

  // boundary conditons
  DomainGhostBC m_domghostbc;

  //domain of fine grid
  ProblemDomain m_domain;

  //the grids at the current level
  DisjointBoxLayout m_grids;

  // Copier object that knows how to perform exchange operations on 
  // LevelData objects defined on m_grids
  Copier m_exchangeCopier;

  /* DisjointBoxLayout for next coarser level
     this is == NULL if there is no coarser level */
  DisjointBoxLayout m_baseBA;

  // number of components
  int m_ncomp;

  // refinement ratio between this and the next coarser level
  int m_refRatio;

  //mesh spacing at this level
  Real m_dxLevel;

  //mesh spacing at next coarser level
  Real m_dxCrse;

  /*
    coarse-fine information between this level
    and the next coarser level
  */
  QuadCFInterp m_quadCFI;

  ///
  /* bottom smoother object
   */
  BaseBottomSmoother* m_bottom_smoother_ptr;

  //has full define function been called?
  bool m_isDefined;

  //has bc been defined?
  bool m_isBCDefined;

  //is inhomogeneous CF interpolation possible?
  bool m_ihcfiEnabled;

  LayoutData<CFIVS> m_loCFIVS[SpaceDim];
  LayoutData<CFIVS> m_hiCFIVS[SpaceDim];

private:
  //internally useful functions
  void
  levelGSRB(
            LevelData<FArrayBox>& a_phi, 
            const LevelData<FArrayBox>& a_rhs);
  void
  interpOnIVS(
              LevelData<FArrayBox>& a_phif, 
              const FArrayBox& a_phistar, 
              const DataIndex& dFine,
              const int a_idir,
              const Side::LoHiSide a_hiorlo,
              const IntVectSet& a_interpIVS);

  void clearMemory();

  void setDefaultValues();

};

#endif
