/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _AMR_H_
#define _AMR_H_

#include <iostream>
#include <string>

#include "REAL.H"
#include "Vector.H"
#include "AMRLevel.H"
#include "AMRLevelFactory.H"
#include "BRMeshRefine.H"
#include "ProblemDomain.H"
#include "Box.H"
#include "HDF5.H"

#include <time.h>

/// framework for Berger-Oliger timestepping for AMR
/**
   This class is a framework for Berger-Oliger timestepping for
   adaptive mesh refinement of time-dependent problems.  It is
   applicable to both hyperbolic and parabolic problems.  It
   represents a hierarchy of levels of refinement as a collection of
   {\tt AMRLevel}s.  
   The usage pattern is this: \\
   1. Call define to define the stuff that does not change throughout 
   the run (maxlevel, refinement ratios, domain, and operator) \\
   1.5. Modify any parameters you like (blocking factor and so forth
   using access functions.\\
   2. Call any one of the three setup functions so AMR can set up all its
   internal data structures.\\
   3. Call run to run the calculation.\\
   4. Call conclude to output (to stdout) how many cells were updated 
   and so forth.
  */
class AMR
{
public:
  ///{\bf constructors, destructors and defines}

  ///
  /**
     Default constructor.  Use must subsequently call a define() function
     and a setUp function.
  */
  AMR();

  ///
  /**
     Destructor.
  */
  ~AMR();

  ///
  /**
     Defines this object.   User must call a setup function
     before running.
     {\bf Arguments:}\\

     a_prob_domain (not modified): problem domain on the base level.\\

     a_ref_ratios (not modified): refinement ratios.  There
     must be at least a_max_level+1 elements,  or an error will 
     result. Element   zero is the base level.\\

     a_max_level (not modified): the maximum level allowed, where the
     base level is zero.  There will be a total of a_max_level+1
     levels, since level zero and level a_max_level will both
     exist.\\

     a_amrlevel_ptr (not modified): pointer to a physics class factory 
     object.
     The object it points to is used to construct the collection of
     AMRLevels in this AMR as objects of the physics class
     type.  It can be destructed after this define()
     returns.  
  */
  void define(int a_max_level, 
              const Vector<int>& a_ref_ratios,
              const Box& a_prob_domain,
              const AMRLevelFactory* const a_amrLevelFact);


  ///
  /**
     Defines this object.   User must call a setup function
     before running.
     {\bf Arguments:}\\

     a_prob_domain (not modified): problem domain on the base level.\\

     a_ref_ratios (not modified): refinement ratios.  There
     must be at least a_max_level+1 elements,  or an error will 
     result. Element   zero is the base level.\\

     a_max_level (not modified): the maximum level allowed, where the
     base level is zero.  There will be a total of a_max_level+1
     levels, since level zero and level a_max_level will both
     exist.\\

     a_amrlevel_ptr (not modified): pointer to a physics class factory 
     object.
     The object it points to is used to construct the collection of
     AMRLevels in this AMR as objects of the physics class
     type.  It can be destructed after this define()
     returns.  
  */
  void define(int a_max_level, 
              const Vector<int>& a_ref_ratios,
              const ProblemDomain& a_prob_domain,
              const AMRLevelFactory* const a_amrLevelFact);


  ///{\bf Setup functions}


  ///
  /**
     Has a define function been called?  Lots of these
     functions will assert fail if not.
   */
  bool isDefined() const;

  ///
  /**
     Has a setUp function been called?  If not,
     you can't call run
   */
  bool isSetUp() const;

#ifdef HDF5
  ///
  /**
     Sets up this object from checkpointed data.  User must have
     previosuly called define.
     Need to call this or setupForNewAMRRun or setupforFixedHierarchyRun 
     before you run.
  */
  void
  setupForRestart(HDF5Handle& a_handle);
#endif

  ///
  /**
     Sets up this object for cold start.  User must have previously
     called define().
     Need to call this or setupForRestart or setupforFixedHierarchyRun 
     before you run.
   */
  void
  setupForNewAMRRun();

  /**
     This sets the hierarchy and sets regrid_intervals to -1
     (turns off regridding).  If you want to keep regridding
     on, reset regridIntervals after this call;
   */
  void
  setupForFixedHierarchyRun(const Vector<Vector<Box> >& a_amr_grids,
                            int a_proper_nest = 1);

  ///
  /**
     You should call this last.  It writes the last
     checkpoint file and tells you how many cells you updated and 
     all that.
   */
  void
  conclude() const;

  ///
  /**
     Runs the calculation.  User must have previously called
     both the define function and a setup function.
  */
  void
  run(Real a_max_time, int a_max_step);

  ///{\bf Administrative functions}

  ///
  /**
     Sets the plot file prefix.
     Should be called after define
     and before setup.
  */
  void
  plotPrefix(const std::string& a_plotfile_prefix);

  ///
  /**
     Sets the checkpoint file prefix.
     Should be called after define
     and before setup.
  */
  void
  checkpointPrefix(const std::string& a_checkpointfile_prefix);

  ///
  /**
     Sets the interval to write plot files, in terms of the base level
     time step.
  */
  void
  plotInterval(int a_plot_interval);

  ///
  /**
     Sets the interval to write checkpoint files, in terms of the base level
     time step.
  */
  void
  checkpointInterval(int a_checkpoint_interval);

  ///
  /**
     Set the maximum grid size.  Should be called after define
     and before setup.
   */
  void
  maxGridSize(int a_max_grid_size);

  ///
  /**
     Set the factor by which the current dt must exceed the new (max) dt
     for time subcycling to occur (i.e., reduction of the current dt by
     powers of 2).
   */
  void
  dtToleranceFactor(Real a_dt_tolerance_factor);

  ///
  /**
     Set the fill ratio for meshrefinement.  Should be called after define
     and before setup.
   */
  void
  fillRatio(Real a_fillRat);

  ///
  /**
     Set the blocking factor for meshrefinement.  Should be called after define
     and before setup.
   */
  void 
  blockFactor(int a_blockFactor);

  ///
  /**
     Set the buffering for the meshrefinement. Should be called after define
     and before setup.
  */
  void 
  gridBufferSize(int a_grid_buffer_size);

  int
  maxGridSize() const;

  ///
  /**
     Sets verbosity level to a_verbosity. 

     -inf to 0: print nothing. \\
     1: prints a message every coarse time step. \\
     2: also prints a message when regridding and writing files. \\
     3: also prints function trace, and some more inf during file I/O. \\
     4 to inf: also prints list of boxes and processor maps during regrid.

     This should be OK to call any time
     after define and before run.  
  */
  void
  verbosity (int a_verbosity);

  ///
  /**
     Sets the regridding intervals.  This should be OK to call any time
     after define and before run.
   */
  void 
  regridIntervals(const Vector<int>& a_regridIntervals);

  ///
  /**
     Returns current verbosity level.  

     -inf to 0: print nothing. \\
     1: prints a message every coarse time step. \\
     2: also prints a message when regridding and writing files. \\
     3: also prints function trace, and some more inf during file I/O. \\
     4 to inf: also prints list of boxes and processor maps during regrid.
  */
  int
  verbosity () const;

  /// 
  /** Set maximum factor by which a timestep can grow. */
  void
  maxDtGrow(Real a_dtGrowFactor);

  /// 
  /** Returns current maximum factor by which a timestep can grow. */
  Real
  maxDtGrow() const;

  /// 
  /** Set a fixed timestep. */
  void
  fixedDt(Real a_dt);

  /// 
  /** Returns the fixed timestep. */
  Real
  fixedDt() const;
 
  Vector<AMRLevel*> getAMRLevels();

  Real getCurrentTime() const;
protected:
  // advance by dt on this level and all finer levels and return the
  // number of steps left - given the number of steps left on entry.
  int
  timeStep(int a_level, int a_stepsLeft, bool timeBoundary);

  //intenal use only 
  void
  clearMemory();
  // make new grids.
  void
  regrid(int a_base_level);

  // make initial grids.
  // not an accurate name, since there are no grids to re.
  void
  initialGrid();

  void
  writePlotFile() const;

  void
  writeCheckpointFile() const;

  // computes maximum stable time step given the maximum stable time
  // step on the individual levels.
  void
  assignDt();

  // whether regridding should be done now.
  bool
  needToRegrid(int a_level, int a_numStepsLeft) const;

  void
  makeBaseLevelMesh (Vector<Box>& a_grids) const;

  void setDefaultValues();

protected:
  int m_blockFactor;
  Real m_fillRatio;
  int m_max_level;
  int m_finest_level_old;
  int m_finest_level;
  int m_checkpoint_interval;
  int m_plot_interval;
  int m_max_grid_size;
  Real m_dt_tolerance_factor;
  Real m_fixedDt;

  bool m_isDefined;
  bool m_isSetUp;
  Vector<AMRLevel*> m_amrlevels;
  Vector<int> m_ref_ratios;
  Vector<int> m_reduction_factor;
  Vector<int> m_regrid_intervals;
  Real m_dt_base;
  // New (maximum) dt
  Vector<Real> m_dt_new;
  // Current dt
  Vector<Real> m_dt_cur;
  Real m_maxDtGrow;

  BRMeshRefine m_mesh_refine;
  bool m_use_meshrefine;

  Vector<Vector<Box> > m_amr_grids;

  int m_cur_step;
  int m_restart_step;
  int m_lastcheck_step;

  Real m_cur_time;
  Vector<int> m_steps_since_regrid;
  Vector<long> m_cell_updates;

  std::string m_plotfile_prefix;
  std::string m_checkpointfile_prefix;

  int m_verbosity;
};
#endif
