/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//  -*- Mode: C++; Modified: "Tue 02 Feb 1999 14:29:43 by graves"; -*- 
#include "LayoutIterator.H"
#include "CFStencil.H"
#include "LayoutIterator.H"
#include "DataIterator.H"
#include "ProblemDomain.H"

bool 
CFStencil::isDefined() const
{ 
    return m_isDefined;
}

/**
   are there any interpolation points?  \\
   returns false if so.*/
bool 
CFStencil::isEmpty() const
{
    assert(m_isDefined);
    return m_fineIVS.isEmpty();
} 

/** get fine intvects which need to be interpolated  \\
    This will be empty if isEmpty() returns true*/
const IntVectSet& 
CFStencil::getFineIVS() const 
{
    assert(m_isDefined);
    return m_fineIVS;
}

/** get coarse intvects that underly fiIVS.
    This will be empty if isEmpty() returns true*/
const IntVectSet& 
CFStencil::getCoarIVS() const 
{
    assert(m_isDefined);
    return m_coarIVS;
}

void  
CFStencil::clear()
{
    m_fineIVS.define();
    m_coarIVS.define();
    setDefaultValues();
}
void  
CFStencil::setDefaultValues()
{
    m_isDefined = false;
    m_direction = -777;
}
CFStencil::CFStencil()
{
    setDefaultValues();
}
CFStencil::CFStencil(const Box& a_b, int a_nComp)
{
  setDefaultValues();
  define(a_b, a_nComp);
    
}
void
CFStencil::define(const Box& a_b, int a_nComp)
{
  clear();
}

CFStencil::~CFStencil()
{
    clear();
}

CFStencil::CFStencil(
    const Box& a_fineDomain, 
    const Box& a_grid, 
    const DisjointBoxLayout& a_fineBoxes, 
    const DisjointBoxLayout& a_coarBoxes, 
    int a_refRatio, 
    int a_direction, 
    Side::LoHiSide a_hiorlo)
{
  ProblemDomain physDomain(a_fineDomain);
  setDefaultValues();
  define(physDomain,  a_grid, a_fineBoxes, a_coarBoxes, 
         a_refRatio, a_direction, a_hiorlo);
}


CFStencil::CFStencil(
                     const ProblemDomain& a_fineDomain,
                     const Box& a_grid, 
                     const DisjointBoxLayout& a_fineBoxes, 
                     const DisjointBoxLayout& a_coarBoxes, 
                     int a_refRatio, 
                     int a_direction, 
                     Side::LoHiSide a_hiorlo)
{
  setDefaultValues();
  define(a_fineDomain, a_grid, a_fineBoxes, a_coarBoxes, 
         a_refRatio, a_direction, a_hiorlo);
}

void
CFStencil::define(
                  const Box& a_fineDomain, 
                  const Box& a_grid, 
                  const DisjointBoxLayout& a_fineBoxes, 
                  const DisjointBoxLayout& a_coarBoxes, 
                  int a_refRatio, 
                  int a_direction, 
                  Side::LoHiSide a_hiorlo)
{
  ProblemDomain physDomain(a_fineDomain);
  define(physDomain, a_grid, a_fineBoxes, a_coarBoxes, a_refRatio, 
         a_direction, a_hiorlo);
}

void
CFStencil::define(
                  const ProblemDomain& a_fineDomain,
                  const Box& a_grid, 
                  const DisjointBoxLayout& a_fineBoxes, 
                  const DisjointBoxLayout& a_coarBoxes, 
                  int a_refRatio, 
                  int a_direction, 
                  Side::LoHiSide a_hiorlo)
{
  m_isDefined = true;
  assert(a_refRatio >= 1);
  assert(a_direction >= 0);
  assert(a_direction < SpaceDim);
  assert((a_hiorlo == Side::Lo) || 
         (a_hiorlo == Side::Hi));
  
  
  assert(!a_fineDomain.isEmpty());    
  
  //set internal vars.  most of these are kept around
  //just to keep the class from having an identity crisis.
  m_direction = a_direction;
  m_hiorlo =  a_hiorlo;
  
  Box finebox = a_grid;
  //this things are guaranteed by proper nesting.
  ProblemDomain coardom= coarsen(a_fineDomain, a_refRatio);
  
  //compute intvectset of all points on fine grid that
  //need to be interpolated
  
  //shift direction
  int hilo = sign(a_hiorlo);
  
  //create fine stencil
  Box edgebox;
  assert((hilo ==1) || (hilo == -1));
  if(hilo == -1)
    {
      edgebox = adjCellLo(finebox,m_direction,1);
    }
  else
    {
      edgebox = adjCellHi(finebox,m_direction,1);
    }
  edgebox = a_fineDomain & edgebox;
  if(!edgebox.isEmpty())
    {
      Box periodicTestBox(a_fineDomain.domainBox());
      if (a_fineDomain.isPeriodic())
        {
          for (int idir=0; idir<SpaceDim; idir++)
            {
              if (a_fineDomain.isPeriodic(idir))
                {
                  periodicTestBox.grow(idir,-1);
                }
            }
        }

      m_fineIVS.define(edgebox);
      LayoutIterator lit = a_fineBoxes.layoutIterator();
      for(lit.reset(); lit.ok(); ++lit)
        {
          m_fineIVS -= a_fineBoxes[lit()];
          // if periodic, also need to subtract periodic images
          // only do this IF we're periodic _and_ both boxes
          // adjoin the domain box boundary somewhere
          if (a_fineDomain.isPeriodic() && !periodicTestBox.contains(edgebox) 
              && !periodicTestBox.contains(a_fineBoxes[lit()]))
            {
              ShiftIterator shiftIt = a_fineDomain.shiftIterator();
              IntVect shiftMult(a_fineDomain.domainBox().size());
              Box shiftedBox(a_fineBoxes[lit()]);
              for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
                {
                  IntVect shiftVect = shiftMult*shiftIt();
                  shiftedBox.shift(shiftVect);
                  m_fineIVS -= shiftedBox;
                  shiftedBox.shift(-shiftVect);
                } // end loop over periodic shift directions
            } // end if periodic
        }
    }
  
  //ivs where all coarse slopes are defined
  //== coarsened fine ivs
  m_coarIVS.define(m_fineIVS);
  m_coarIVS.coarsen(a_refRatio);
  // this is a trick to get around the lack of a IntVectSet intersection
  // operator which works with a ProblemDomain
  Box domainIntersectBox = m_coarIVS.minBox();
  domainIntersectBox = coardom & domainIntersectBox;
  m_coarIVS &= domainIntersectBox;
}


CFStencil&
CFStencil::operator=(const CFStencil& a_ecfsIn)
{
    if(&a_ecfsIn != this)
    {
        m_direction= a_ecfsIn.m_direction;                 
        m_hiorlo = a_ecfsIn.m_hiorlo;
        m_dataIndex = a_ecfsIn.m_dataIndex;
        m_fineIVS = a_ecfsIn.m_fineIVS;
        m_coarIVS= a_ecfsIn.m_coarIVS;  
    }
    return *this;
}
