/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _Deriv_Stencil_H_
#define _Deriv_Stencil_H_

#include <iostream>
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include "REAL.H"
#include "IntVect.H"
#include "Vector.H"

///class to encapsulate the operations to create derivs on irreg stencils
/**
   DerivStencil is meant to be used to encapsulate the information
   necessary to take finite difference derivatives at a point in 
   space.  Every point in the stencil has a weight.  You add them
   (the boxarrayindex and the weight) at the same time and you can manipulate
   the weights enmasse by real number operations.  Stencils may not
   interact with each other with the same sort of arithmetic because
   that would bring up issues as to what to do when there is 
   incomplete intersection between the stencils.

 */
class DerivStencil
{
public:
  ///
  void define();

  ///
  /** return true if  any define function been called.  
   */
  bool isDefined() const
  {return isdefined;}

  ///
  /**
     default constructor;  creates empty vectors
  */
  DerivStencil();

  ///
  /**
     copy constructor;  sets *this = a_dsin
  */
  DerivStencil(const DerivStencil& a_dsin);

  ///
  ~DerivStencil();

  ///
  /**
     make derivstencil empty
  */
  void clear();

  ///
  /**
     return length of vectors
  */
  int size() const;

  ///
  /**
     get iv at ivec
  */
  const IntVect& getIndex(int a_ivec) const;

  ///
  /**
     get weight at ivec
  */
  const Real& getWeight(int a_ivec) const;

  ///
  /**
     add another set if the IntVect is not in the 
     stencil already.  \\
     **Add the weight to the existing weight otherwise**
     */
  void accumulate(const IntVect& a_iv,  Real a_weight);

  ///
  /**
     assignment operator
  */
  const DerivStencil& operator=(const DerivStencil& a_dsin);

  ///
  /**
     Multiply each weight by a_facin
     does nothing if vectors are of zero length
  */
  const DerivStencil& operator*=(Real a_facin);

  ///
  /**
     divide each weight by a_denom
     does nothing if vectors are of zero length.
     Be advised ---
     This function does no checking to see if a_denom
     is close to zero.  
  */
  const DerivStencil& operator/=(Real a_denom);

  ///
  /**
     add a_facin to each weight
     does nothing if vectors are of zero length
  */
  const DerivStencil& operator+=(Real a_facin);

  ///
  /**
     subtract a_facin from each weight
     does nothing if vectors are of zero length
  */
  const DerivStencil& operator-=(Real a_facin);

protected:
  ///
  Vector<IntVect> m_vectIV;
  ///
  Vector<Real>          m_vectWgt;
  ///
  bool isdefined;
};



#endif
