/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "QuadCFInterp.H"
#include "LayoutIterator.H"
#include "DataIterator.H"

/***********************/
// default constructor
/***********************/
void
QuadCFInterp::clear()
{
  m_isDefined = false;
  m_level = -1;
  m_dxFine = -1;
}

QuadCFInterp::QuadCFInterp()
{
  clear();
}

/***********************/
/***********************/
QuadCFInterp::QuadCFInterp(
                           const DisjointBoxLayout& a_fineBoxes, 
                           const DisjointBoxLayout* a_coarBoxes, 
                           Real  a_dxFine,
                           int a_refRatio, int a_nComp,
                           const Box& a_domf)
{
  ProblemDomain fineProbDomain(a_domf);
  define(a_fineBoxes,a_coarBoxes, a_dxFine,a_refRatio,a_nComp, fineProbDomain);
}


/***********************/
/***********************/
QuadCFInterp::QuadCFInterp(
                           const DisjointBoxLayout& a_fineBoxes, 
                           const DisjointBoxLayout* a_coarBoxes, 
                           Real  a_dxFine,
                           int a_refRatio, int a_nComp,
                           const ProblemDomain& a_domf)
{
  define(a_fineBoxes,a_coarBoxes, a_dxFine,a_refRatio,a_nComp, a_domf);
}

/***********************/
/***********************/

void
QuadCFInterp::define(
                     const DisjointBoxLayout& a_fineBoxes,
                     const DisjointBoxLayout* a_coarBoxesPtr,
                     Real  a_dxLevel, 
                     int a_refRatio, int a_nComp,
                     const Box& a_domf)
{
  ProblemDomain fineProbDomain(a_domf);
  define(a_fineBoxes, a_coarBoxesPtr, a_dxLevel, a_refRatio, a_nComp,
         a_domf);

}


/***********************/
/***********************/

void
QuadCFInterp::define(
                     const DisjointBoxLayout& a_fineBoxes,
                     const DisjointBoxLayout* a_coarBoxesPtr,
                     Real  a_dxLevel, 
                     int a_refRatio, int a_nComp,
                     const ProblemDomain& a_domf)
{
  clear();
  m_isDefined = true;
  assert(a_nComp > 0);

  assert (!a_domf.isEmpty());
  // consistency check
  assert (a_fineBoxes.checkPeriodic(a_domf));

  m_dxFine = a_dxLevel;
  m_refRatio = a_refRatio;
  m_nComp = a_nComp;
  m_inputFineLayout = a_fineBoxes;
  if(a_coarBoxesPtr == NULL)
    m_level = 0;
  else
    m_level = 1;

  if(m_level > 0)
    {
      // (DFM) only check for valid refRatio if a coarser level exists
      assert(a_refRatio >= 1);
      const DisjointBoxLayout& coarBoxes = *a_coarBoxesPtr;
      m_inputCoarLayout = coarBoxes;
      assert (coarBoxes.checkPeriodic(coarsen(a_domf,a_refRatio)));

      for(int idir = 0; idir < SpaceDim; idir++)
        {
          m_loQCFS[idir].define(a_fineBoxes);
          m_hiQCFS[idir].define(a_fineBoxes);
          m_loCoarBoxes[idir].deepCopy(a_fineBoxes);
          m_hiCoarBoxes[idir].deepCopy(a_fineBoxes);
        }

      //locoarboxes and hicoarboxes are now open
      //and have same processor mapping as a_fineboxes
      //make boxes for coarse buffers
      LayoutIterator lit = a_fineBoxes.layoutIterator();
      for(lit.begin(); lit.ok(); ++lit)
        {
          const Box& fineBox = a_fineBoxes[lit()];
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              Box coarBox   = coarsen(fineBox, m_refRatio);
              Box coarBoxHi = adjCellHi(coarBox,idir,1); 
              Box coarBoxLo = adjCellLo(coarBox,idir,1); 
              //give em a ghost cell on non-idir sides
              coarBoxLo.grow(1);
              coarBoxLo.grow(idir, -1);
              coarBoxHi.grow(1);
              coarBoxHi.grow(idir, -1);
              
              m_loCoarBoxes[idir][lit()] = coarBoxLo;
              m_hiCoarBoxes[idir][lit()] = coarBoxHi;
            }
        }

      for(int idir = 0; idir < SpaceDim; idir++)
        {
          m_loCoarBoxes[idir].close();
          m_hiCoarBoxes[idir].close();
          m_loCoarBuffer[idir].define(m_loCoarBoxes[idir],m_nComp);
          m_loCoarCopiers[idir].define(coarBoxes, m_loCoarBoxes[idir]);
          m_hiCoarBuffer[idir].define(m_hiCoarBoxes[idir],m_nComp);
          m_hiCoarCopiers[idir].define(coarBoxes, m_hiCoarBoxes[idir]);
        }

      //make cfstencils and boxes for coarse buffers
      DataIterator dit = a_fineBoxes.dataIterator();
      for(dit.begin(); dit.ok(); ++dit)
        {
          const Box& fineBox = a_fineBoxes[dit()];
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              //low side cfstencil
              m_loQCFS[idir][dit()].define(a_domf,
                                           fineBox,
                                           a_fineBoxes,
                                           coarBoxes,
                                           a_refRatio,
                                           idir,
                                           Side::Lo);
              //high side cfstencil
              m_hiQCFS[idir][dit()].define(a_domf,
                                           fineBox,
                                           a_fineBoxes,
                                           coarBoxes,
                                           a_refRatio,
                                           idir,
                                           Side::Hi);
              
            }
        }
    }
}


/***********************/
//  apply coarse-fine boundary conditions -- assume that phi grids
//  are grown by one
/***********************/
void
QuadCFInterp::coarseFineInterp(BaseFab<Real> & a_phif,
                               const BaseFab<Real> & a_phic, 
                               const QuadCFStencil& a_qcfs,
                               const Side::LoHiSide a_hiorlo,
                               const int a_idir,
                               const Interval& a_variables) const
{
  assert(isDefined());
  //nothing happens if m_level == 0
  if(m_level > 0)
    {
      if(!a_qcfs.isEmpty())
        {
          BaseFab<Real>  phistar;

          //first find extended value phistar
          //includes finding slopes of coarse solution bar
          getPhiStar(phistar, a_phic, a_qcfs, a_hiorlo, a_idir, a_variables);

          //given phistar, interpolate on fine ivs
          interpOnIVS(a_phif, phistar, a_qcfs, a_hiorlo, a_idir, a_variables);
        }
    }
}

/***********************/
//get extended phi (lives next to interpivs)
/***********************/
void 
QuadCFInterp::getPhiStar(BaseFab<Real> & a_phistar, 
                         const BaseFab<Real> & a_phic,
                         const QuadCFStencil& a_qcfs,
                         const Side::LoHiSide a_hiorlo,
                         const int a_idir, 
                         const Interval& a_variables) const
{
    
  assert(isDefined());
  assert(a_qcfs.isDefined());
  Real dxf = m_dxFine;
  Real dxc = m_refRatio*dxf;

  //if cfsten is empty, nothing to interpolate.
  if(!a_qcfs.isEmpty())
    {
      //if there IS something to interpolate, the level ident 
      //had better be greater than zero.  Otherwise a null
      //was sent in as coarse grids on construction
      assert(m_level > 0);
      const IntVectSet& interp_ivs = a_qcfs.getFineIVS();
      const IntVectSet& coarsl_ivs = a_qcfs.getCoarIVS();
      Box coarinterpbox = coarsl_ivs.minBox();
      int ncomp = a_phic.nComp();
      assert(ncomp == m_nComp);
      assert(a_phic.box().contains((coarinterpbox)));

      // allocate phistar here
      int ihilo = sign(a_hiorlo);
      Box phistarbox = interp_ivs.minBox();      
      phistarbox.shift(a_idir, ihilo);      
      a_phistar.resize(phistarbox, ncomp);


      for(int ivar = a_variables.begin(); ivar <= a_variables.end(); ivar++)
        {
          //phi = phino + slope*x + half*x*x*curvature 
          BaseFab<Real>  coarslope(coarinterpbox, SpaceDim);
          BaseFab<Real>  coarcurva(coarinterpbox, SpaceDim);

#if(CH_SPACEDIM == 3)
          //directions != idir,  needed for mixed derivative in 3d
          //add to phi x*y*mixed_derivative
          Vector<int> xydir;
          for(int isldir = 0; isldir < SpaceDim; isldir++)
            if(isldir != a_idir)
              xydir.push_back(isldir);
          BaseFab<Real>  coarmixed(coarinterpbox, 1);
#endif

          coarslope.setVal(0.);

          //first find extended value phistar
          //get slopes of coarse solution
          IVSIterator coar_ivsit(coarsl_ivs);
          for(coar_ivsit.begin(); coar_ivsit.ok(); ++coar_ivsit)
            {
              IntVect coariv = coar_ivsit ();
              for(int isldir = 0; isldir < SpaceDim; isldir++)
                {
                  if(isldir == a_idir)
                    {
                      coarslope(coariv, isldir) = 0.0;
                      coarcurva(coariv, isldir) = 0.0;
                    }
                  else
                    {
                      coarslope(coariv, isldir) = 
                        a_qcfs.computeFirstDerivative(a_phic, isldir, 
                                                      ivar, coariv, dxc);
                      coarcurva(coariv, isldir) = 
                        a_qcfs.computeSecondDerivative(a_phic, isldir, 
                                                       ivar, coariv, dxc);
                    } //end if(isldir == idir)
                } // end loop over slope dir
#if(CH_SPACEDIM == 3)
              coarmixed(coariv) = 
                a_qcfs.computeMixedDerivative(a_phic, ivar, coariv, dxc);
#endif
            } //end loop over coarse intvects
                
          IVSIterator fine_ivsit(interp_ivs);
          for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit)
            {
              IntVect ivf = fine_ivsit();
              IntVect ivc = coarsen(ivf,m_refRatio);
              IntVect ivstar = ivf;
              ivstar.shift(a_idir, ihilo);
              a_phistar(ivstar, ivar)  = a_phic(ivc,ivar);

              for(int isldir = 0; isldir < SpaceDim; isldir++)
                {
                  if(isldir != a_idir)
                    {
                      int jf = ivf[isldir];
                      int jc = ivc[isldir];
                      Real xf = (jf+0.5)*dxf;
                      Real xc = (jc+0.5)*dxc;
                      Real x = xf-xc;
                      a_phistar(ivstar, ivar) += 
                        (      x*coarslope(ivc, isldir)) + 
                        (0.5*x*x*coarcurva(ivc, isldir));
                    }//end if(isldir != idir)
                } //end loop over slope directions
#if(CH_SPACEDIM==3)
              //add in mixed derivative component
              int dif = ivf[xydir[0]];
              int dic = ivc[xydir[0]];
              int djf = ivf[xydir[1]];
              int djc = ivc[xydir[1]];

              Real xf = (dif+0.5)*dxf;
              Real xc = (dic+0.5)*dxc;
              Real yf = (djf+0.5)*dxf;
              Real yc = (djc+0.5)*dxc;
              Real x = xf-xc;
              Real y = yf-yc;
              a_phistar(ivstar, ivar) += x*y*coarmixed(ivc);
#endif                
            } //end loop over fine intvects
        } //end loop over variables
    } //end if(level>0 && !hocfs.isempty())
} //end function getphistar

/***********************/
/***********************/
bool 
QuadCFInterp::isDefined() const
{
  return m_isDefined;
}

void 
QuadCFInterp::interpOnIVS(BaseFab<Real> & a_phif, 
                          const BaseFab<Real> & a_phistar, 
                          const QuadCFStencil& a_qcfs,
                          const Side::LoHiSide a_hiorlo,
                          const int a_idir,
                          const Interval& a_variables) const
{
  assert(isDefined());
  assert(a_qcfs.isDefined());

  //if cfsten is empty, nothing to interpolate.
  if(!a_qcfs.isEmpty())
    {
      //if there IS something to interpolate, the level ident 
      //had better be greater than zero.  Otherwise a null
      //was sent in as coarse grids on construction
      assert(m_level > 0);
      const IntVectSet& interp_ivs = a_qcfs.getFineIVS();
      IVSIterator fine_ivsit(interp_ivs);
      assert(a_phistar.nComp() == a_phif.nComp());
      assert(a_phistar.nComp() == m_nComp);
        
      int nref = m_refRatio;
      int ihilo = sign(a_hiorlo);
      for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit)
        {
          IntVect ivf = fine_ivsit();
          // quadratic interpolation 
          for(int ivar = a_variables.begin(); ivar <= a_variables.end(); ivar++)
            {
              Real pa =    a_phif(ivf -2*ihilo*BASISV(a_idir), ivar);
              Real pb =    a_phif(ivf -  ihilo*BASISV(a_idir), ivar);
              Real ps = a_phistar(ivf +  ihilo*BASISV(a_idir), ivar);
              //phi = ax**2 + bx + c, x = 0 at pa
              Real h = m_dxFine;
              Real a = (2./h/h)*(2.*ps + pa*(nref+1.0) -pb*(nref+3.0))/
                (nref*nref + 4*nref + 3.0);
              Real b = (pb-pa)/h - a*h;
              Real c = pa;
              Real x = 2.*h;
              a_phif(ivf,ivar) = a*x*x + b*x + c;
            } //end loop over components
        } //end loop over fine intvects
    } //end if(level>0 && !oscfs.isempty())
} //end function interponivs


/***********************/
//  apply coarse-fine boundary conditions -- assume that phi grids
//  are grown by one
/***********************/
void
QuadCFInterp::coarseFineInterp(LevelData<FArrayBox>& a_phif,
                               const LevelData<FArrayBox>& a_phic)
{
  assert(isDefined());
  assert(a_phif.nComp() == m_nComp);
  assert(a_phic.nComp() == m_nComp);
  assert(a_phif.ghostVect() >= IntVect::TheUnitVector());
  assert(a_phic.boxLayout() == m_inputCoarLayout);
  assert(a_phif.boxLayout() == m_inputFineLayout);

  Interval variables = a_phic.interval();

  if(m_level > 0)
    {
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          //put coarse data into buffers
          a_phic.copyTo(a_phic.interval(),
                        m_loCoarBuffer[idir],
                        m_loCoarBuffer[idir].interval(),
                        m_loCoarCopiers[idir]);
          a_phic.copyTo(a_phic.interval(),
                        m_hiCoarBuffer[idir],
                        m_hiCoarBuffer[idir].interval(),
                        m_hiCoarCopiers[idir]);

          DataIterator ditFine = a_phif.dataIterator();
          for(ditFine.begin(); ditFine.ok(); ++ditFine)
            {
              DataIndex datIndGlo =ditFine();
              BaseFab<Real> & phif = a_phif[datIndGlo];
              //lo side cfinterp
              //recall that buffers have fine processor mapping
              {
                const BaseFab<Real> & loPhiC = m_loCoarBuffer[idir][datIndGlo];
                const QuadCFStencil& loQCFS = m_loQCFS[idir][datIndGlo];
                coarseFineInterp(phif,loPhiC,loQCFS, Side::Lo, idir, variables);
              }
              //hi side cfinterp
              {
                const BaseFab<Real> & hiPhiC = m_hiCoarBuffer[idir][datIndGlo];
                const QuadCFStencil& hiQCFS = m_hiQCFS[idir][datIndGlo];
                coarseFineInterp(phif,hiPhiC, hiQCFS, Side::Hi, idir, variables);
              }
            }//end iteration over boxes in fine grid
        } //end iteration over directions
    }
}
/***********************/
/***********************/
QuadCFInterp::~QuadCFInterp() 
{
  clear();
}

