/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "DataIterator.H"
#include "IntVect.H"
#include "Copier.H"
#include "MayDay.H"
#include "LayoutIterator.H"
#include "SPMD.H"

using std::ostream;

Pool Copier::s_motionItemPool(sizeof(MotionItem), "Copier::MotionItem");

Copier::Copier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest)
{
  const ProblemDomain& domain = a_level.physDomain();
  define(a_level, a_dest, domain, IntVect::TheZeroVector());
}

Copier::Copier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest,
               const ProblemDomain& a_domain)
{
  define(a_level, a_dest, a_domain, IntVect::TheZeroVector());
}


Copier::Copier(const DisjointBoxLayout& a_level,
               const BoxLayout& a_dest,
               const IntVect& a_destGhost)
{
  const ProblemDomain domain = a_level.physDomain();
  define(a_level, a_dest, domain, a_destGhost);
}

Copier::Copier(const DisjointBoxLayout& a_level,
               const BoxLayout& a_dest,
               const ProblemDomain& a_domain,
               const IntVect& a_destGhost)
{
  define(a_level, a_dest, a_domain, a_destGhost);
}


void Copier::clear()
{
  for(unsigned int i=0; i<m_localMotionPlan.size(); ++i)
    {
      s_motionItemPool.returnPtr( m_localMotionPlan[i] );
    }
  for(unsigned int i=0; i<m_fromMotionPlan.size(); ++i)
    {
      s_motionItemPool.returnPtr( m_fromMotionPlan[i] );
    }
  for(unsigned int i=0; i<m_toMotionPlan.size(); ++i)
    {
      s_motionItemPool.returnPtr( m_toMotionPlan[i] );
    }

  m_localMotionPlan.resize(0);
  m_fromMotionPlan.resize(0);
  m_toMotionPlan.resize(0);
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest)
{
  const ProblemDomain domain = a_level.physDomain();
  define(a_level, a_dest, domain, IntVect::TheZeroVector());
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest,
                    const ProblemDomain& a_domain)
{
  define(a_level, a_dest, a_domain, IntVect::TheZeroVector());
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest,
                    const IntVect& a_ghost)
{
  const ProblemDomain domain = a_level.physDomain();
  define (a_level, a_dest, domain, a_ghost);
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest,
                    const ProblemDomain& a_domain,
                    const IntVect& a_ghost)
{
  assert(a_level.isClosed());
  assert(a_dest.isClosed());
  assert(a_level.checkPeriodic(a_domain));

  clear();

  //bool self = a_dest == a_level;  
  const DisjointBoxLayout& level= a_level;
  const BoxLayout&         dest = a_dest;

  
  unsigned int myprocID = procID();

  for(LayoutIterator to(a_dest.layoutIterator()); to.ok(); ++to)
    {
      Box ghost(dest[to()]);
      ghost.grow(a_ghost);
      unsigned int toProcID = dest.procID(to());
      for(LayoutIterator from(a_level.layoutIterator()); from.ok(); ++from)
        {
          unsigned int fromProcID = level.procID(from());
          // if(self && from() == to()){
          // don't copy onto self valid region when performing exchange.
          // a_dest == a_level is my signal that I'm performing an
          // 'exchange' type operation.

          if(toProcID != myprocID && fromProcID != myprocID)
            {
              // don't keep track of copy operations that don't involve this processor
            }
          else
            {
              const Box& fromBox = level[from()];
              if (ghost.intersectsNotEmpty(fromBox))
                {
                  Box box(ghost);
                  box&=fromBox;
                  MotionItem* item = new (s_motionItemPool.getPtr())
                  MotionItem(DataIndex(from()), DataIndex(to()), box);
                  if(item == NULL)
                    {
                      MayDay::Error("Out of Memory in copier::define");
                    }
                  if(toProcID == fromProcID) // local move
                    m_localMotionPlan.push_back(item);
                  else if(fromProcID == myprocID)
                    {
                      item->procID = toProcID;
                      m_fromMotionPlan.push_back(item);
                    }
                  else
                    {
                      item->procID = fromProcID;
                      m_toMotionPlan.push_back(item);
                    }
                }
              // put periodic intersection checking in here
              // (innermost loop, since we've got both boxes already)
              if (a_domain.isPeriodic()) 
                {
                  Box domainBox(a_domain.domainBox());
                  assert (!domainBox.isEmpty());
                  
                  // only do this if ghost box hangs over domain edge 
                  // or if fromBox does
                  if (!domainBox.contains(ghost) || !domainBox.contains(fromBox))
                    {
                      ShiftIterator shiftIt = a_domain.shiftIterator();
                      IntVect shiftMult(domainBox.size());                      

                      for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
                        {
                          IntVect shiftVect(shiftIt()*shiftMult);
                          ghost.shift(shiftVect);
                          if (ghost.intersectsNotEmpty(fromBox))
                            {
                              Box intersectBox(ghost);
                              intersectBox &= fromBox;
                              Box toBox(intersectBox);
                              toBox.shift(-shiftVect);
                              MotionItem* item = new (s_motionItemPool.getPtr())
                                MotionItem(DataIndex(from()), DataIndex(to()),
                                           intersectBox, toBox);
                              if (item == NULL) 
                                { 
                                  MayDay::Error("Out of Memory in copier::define");
                                }
                              if (toProcID == fromProcID) // local move
                                m_localMotionPlan.push_back(item);
                              else if (fromProcID == myprocID)
                                {
                                  item->procID = toProcID;
                                  m_fromMotionPlan.push_back(item);
                                }
                              else
                                {
                                  item->procID = fromProcID;
                                  m_toMotionPlan.push_back(item);
                                }
                              
                            } // end if shifted box intersects
                          ghost.shift(-shiftVect);
                        } // end loop over shift vectors                  
                    } //end if ghost box is not contained in domain
                } // end if periodic
              else if(fromBox.smallEnd(0) > ghost.bigEnd(0))
                {
                  //can skip the rest, since we know that the smallEnd
                  // of all the remaining boxes
                  // are lexigraphically beyond this ghosted box.
                  from.end();
                }
            }
        }
    }
}

int Copier::print() const
{
  std::cout << *this;
  return 0;
}

ostream& operator<<(ostream& os, const Copier& copier)
{
  os << "local("<<procID()<<"): ";
  for(CopyIterator it(copier, CopyIterator::LOCAL); it.ok(); ++it)
    {
      os << it().toRegion;
    }
  os << "\nfrom("<<procID()<<"): ";
  for(CopyIterator it(copier, CopyIterator::FROM); it.ok(); ++it)
    {
      os << it().fromRegion<<"["<<it().procID<<"]";
    }
  os << "\nto("<<procID()<<"): ";
  for(CopyIterator it(copier, CopyIterator::TO); it.ok(); ++it)
    {
      os << it().toRegion<<"["<<it().procID<<"]";
    }
  os<<"\n";
  return os;
}

Copier::~Copier()
{
  clear();
}

