/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_PROBDOMAIN_H
#define CH_PROBDOMAIN_H

#ifndef WRAPPER
#include <iostream>

#include "Vector.H"
#include "IntVect.H"
#include "Box.H"
#include "Misc.H"
#endif

#include "SPACE.H"

/// clase to manage box-shifting used to enforce periodic BC's
/** The ShiftIterator class contains a list of shift vectors necessary
    to enforce periodic boundary conditions (enforced by shifting the 
    Box of each Fab, then copying any overlapping valid cells with
    ghost cells of the shifted Fab.  Depending on the number of periodic 
    directions, the list of shift vectors over which to iterate differs.
*/
class ShiftIterator
{
public:
  /// basic constructor
  ShiftIterator();

  /// 
  ShiftIterator(const bool* a_isPeriodic);

  /// copy constructor
  ShiftIterator(const ShiftIterator& a_shiftIt);

  /// 
  ~ShiftIterator();

  ///
  ShiftIterator& operator=(const ShiftIterator& a_src);

  /// recompute shift vectors based on periodic directions
  void computeShifts(const bool* a_isPeriodic);

  /// returns the current shift unit vector
  inline IntVect operator()() const;
  IntVect i() const {return this->operator()();}

  /// increment to the next shift unit vector
  inline void operator++();
  void incr(){++(*this);}

  ///
  inline bool ok() const;

  /// reset to first shift unit vector
  inline void reset();
  
  /// 
  inline void begin();

  /// 
  /** skip the iterator to the end.  ok() will return false after this
      method is called.
  */
  void end();

private:

  unsigned int m_index;

  Vector<IntVect> m_shift_vectors;

};



//
//@Man:
//@Memo: A class to facilitate interaction with physical boundary conditions
/*@Doc: 

  ProblemDomain is a class which facilitates the application of physical 
  boundary conditions, both periodic and non-periodic.  This class contains 
  much of the functionality of the Box class, since logically the 
  computational domain is generally a Box.  
  
  Intersection with a ProblemDomain object will result in only removing 
  regions which are outside the physical domain in non-periodic directions.
  Regions outside the logical computational domain in periodic directions will
  be treated as ghost cells which can be filled with an exchange() function 
  or through suitable interpolation from a coarser domain.

  Since ProblemDomain will contain a Box, it is a dimension dependent class, 
  so SpaceDim must be defined as either 1, 2, or 3 when compiling.  

  Note that this implementation of ProblemDomain is inherently 
  cell-centered.


*/
class ProblemDomain
{
public:

  /// {\bf Constructors}

  ///
  /**
     The default constructor.  The constructed domain box is empty. 

  */
    ProblemDomain ();

  ///
  /**
     Construct ProblemDomain with a_domBox as computational domain
     This constructor defaults to non-periodic domain
  */
  ProblemDomain(const Box& a_domBox);

  ///
  /**
     Construct ProblemDomain with a_domBox as computational domain.
     a_isPeriodic is a SpaceDim array of bools, true is the physical
     boundary condition is periodic in the coordinate direction
  */
  ProblemDomain(const Box& a_domBox, const bool* a_isPeriodic);

  ///
  /**
     Construct a ProblemDomain.  It is an error if small is
     greater than big.

  */
    ProblemDomain (const IntVect& small,
         const IntVect& big);


  ///
  /**
     Construct a ProblemDomain.  It is an error if small is
     greater than big.
     a_isPeriodic is a SpaceDim array of bools, true is the physical
     boundary condition is periodic in the coordinate direction

  */
    ProblemDomain (const IntVect& small,
                   const IntVect& big,
                   const bool* a_isPeriodic);


  ///
  /**
     Construct ProblemDomain with specified lengths.  It is an error
    if the lengths are negative.

  */
    ProblemDomain (const IntVect& small,
         const int*     vec_len);


  ///
  /**
     Construct ProblemDomain with specified lengths.  It is an error
    if the lengths are negative.
     a_isPeriodic is a SpaceDim array of bools, true is the physical
     boundary condition is periodic in the coordinate direction

  */
    ProblemDomain (const IntVect& small,
                   const int*     vec_len,
                   const bool* a_isPeriodic);


  ///
  /**
     The copy constructor.

  */
    ProblemDomain (const ProblemDomain& a_src);

  /// {\bf Accessors}

  ///
  /** 
      Returns the logical computational domain
  */
  const Box& domainBox() const;
  

  ///
  /**
     returns true if BC is periodic in direction a_dir
  */
  bool isPeriodic(int a_dir) const;

  ///
  /**
     returns true is BC is periodic in _any_ direction
  */
  bool isPeriodic() const;
  
  
  /// 
  /** 
      returns the shiftIterator for this ProblemDomain
  */
  ShiftIterator shiftIterator() const;


  ///
  /**
     Returns true if this ProblemDomain is empty or undefined.     
  */
  bool isEmpty () const;
  
  ///
  /**
     Returns true if argument is contained within this ProblemDomain.  An empty
     ProblemDomain does not contain and is not contained by any ProblemDomain, 
     including itself.    
  */
  bool contains (const IntVect& p) const;
  ///
  /**
	 Returns the periodic image of this IntVect inside of the ProblemDomain.
	 Return true if the domain contains this IntVect, returning the image
	 in 'p', otherwise returns false
  */
  bool image (IntVect& p) const;
  
  ///
  /**
     Returns true if argument is contained within this ProblemDomain.  It 
     is an error if the Box is not cell-centered.  An empty
     ProblemDomain does not contain any Box.
  */
    bool contains (const Box& b) const;
  bool contains_box(const Box& b) const {return contains(b);}

  /// 
  /**
     Returns true if this ProblemDomain and the argument have non-null
     intersections.  It is an error if a_box is not cell-centered.
     An empty ProblemDomain does not intersect any Box.
     This will do nothing in periodic directions (since a periodic domain 
     is infinite in the periodic direction.  If periodic in all dimensions, 
     this will always return true.
  */
  bool intersects (const Box& a_box) const;

  /// 
  /**
     Returns true if this ProblemDomain and the argument have non-null
     intersections.  It is an error if a_box is not cell-centered.
     This routine does not perform the check to see if *this or b are
     empty boxes.  It is the callers responsibility to ensure that
     this never happens.  If you are unsure, the use the .intersects(..) 
     routine.  In periodic directions, will always return true.

*/
    bool intersectsNotEmpty (const Box& a_box) const;


  /// 
  /** returns true if box1 and box2 and any of their periodic 
      images intersect (useful for checking disjointness)
  */ 
  bool intersects(const Box& box1, const Box& box2) const;

  ///
  /** modifies a_box to be the intersection of a_box and the 
      problemDomain
  */
  friend void operator &=(Box& a_box, const ProblemDomain& a_probomain);

  ///
  /**  returns a box which is the interesection of a_box and the 
       problemDomain
  */
  friend Box operator & (const Box& a_box, const ProblemDomain& a_probdomain);

  /// {\bf Modification Functions}

  ///
  /**
    The assignment operator.

  */
  ProblemDomain& operator= (const ProblemDomain& b);

  ///
  /** 
      sets whether BC is periodic in direction a_dir (true == periodic)
  */
  void setPeriodic(int a_dir, bool a_isPeriodic);
  
  ///
  /** grows (or shrinks) the domain box by i in all directions
   */
  inline ProblemDomain& grow (int i);

  ///
  /** returns a ProblemDomain with a domainBox grown by the given amount
   */
  friend inline ProblemDomain grow(const ProblemDomain& pd, 
                                   int i);

  ///
  /**  
       modifies this ProblemDomain by growing the domainBox in each
       direction by the specified amount
  */
  inline ProblemDomain& grow(const IntVect& v);

  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain
     with a DomainBox grown by the given amount.
  */
  friend inline ProblemDomain grow (const ProblemDomain& pd,
                                     const IntVect& v);

  ///
  /** Modifies this ProblemDomain by growing it on the low and high end
      by n_cell cells in direction idir
  */
  inline ProblemDomain& grow(int idir, int n_cell);

  ///
  /** Modifies this ProblemDomain by growing it on the low end by n_cell
      cells in direction idir.
  */
  inline ProblemDomain& growLo(int idir, int n_cell=1);

  ///
  /** Modifies this ProblemDomain by growing it on the high end by n_Cell
      cells in direction idir
  */
  inline ProblemDomain& growHi(int idir, int n_cell=1);
     


  ///
  /**
     Returns the edge-centered Box (in direction dir) defining the low
     side of the argument ProblemDomain.  The output Box will have the given
     length in the given direction.  Directions are zero-based.  It is
     an error if not 0 <= dir < SpaceDim.  The neighbor of an Empty ProblemDomain
     is an Empty Box of the appropriate type.  If dir is a periodic
     direction, will return an empty box.
     
  */
  friend Box bdryLo (const ProblemDomain& a_pd,
                     int        a_dir,
                     int        a_len=1);
  
  /// 
  /**
     Returns the edge-centered Box (in direction dir) defining the
     high side of the argument ProblemDomain.  The return Box will have the
     given length in the given direction.  Directions are zero-based.
     It is an error if not 0 <= dir < SpaceDim.  The neighbor of an
     Empty ProblemDomain is an Empty Box of the appropriate type.  
     If dir is a periodic direction, will return an empty box.
     
  */
  friend Box bdryHi (const ProblemDomain& a_pd,
                     int        a_dir,
                     int        a_len=1);
  
  ///
  /**
     Returns the cell centered Box of the given length adjacent to the
     argument ProblemDomain on the low end along the given coordinate direction.
     The return Box is identical to the argument ProblemDomain in the other
     directions.  The return ProblemDomain and the argument ProblemDomain have an 
     empty intersection.\\

     NOTE: len >= 1. \\

     NOTE: Box retval = adjCellLo(b,dir,len) is equivalent to the
     following set of operations: \\ 

     Box retval(b); \\

     retval.convert(dir,ProblemDomain::CELL); \\

     retval.setrange(dir,retval.smallEnd(dir)-len,len); \\

     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The neighbor of an Empty ProblemDomain is an Empty Box of the
     appropriate type. If dir is a periodic direction, will return
     an empty Box as well.

  */
  friend Box adjCellLo (const ProblemDomain& a_pd,
                        int        a_dir,
                        int        a_len=1);
  
  ///
  /**
     Returns the cell centered Box of the given length adjacent to the
     argument ProblemDomain on the high end along the given coordinate
     direction.  The return Box is identical to the argument ProblemDomain in
     the other directions.  The return Box and the argument ProblemDomain have
     an empty intersection.  \\
     
     NOTE: len >= 1. \\
     
     NOTE: Box retval = adjCellHi(b,dir,len) is equivalent to the
     following set of operations: \\ 
     
     Box retval(b); \\
     
     retval.convert(dir,ProblemDomain::CELL); \\
     
     retval.setrange(dir,retval.bigEnd(dir)+1,len);\\
     
     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The neighbor of an Empty ProblemDomain is an Empty Box of the
     appropriate type.    If dir is a periodic direction, will return
     an empty Box.
     
  */
  friend Box adjCellHi (const ProblemDomain& a_pd,
                        int        a_dir,
                        int        a_len=1);
  
  
  /// intersection functions 
  
  ///
  /**
     Returns the Box that is intersection of this ProblemDomain and the argument
     ProblemDomain.  The Box MUST be cell-centered.
     The intersection of the Empty ProblemDomain and any Box is the Empty
     Box. This operator does nothing in periodic directions (since
     a periodic domain is an infinite domain).
     
  */
  Box operator& (const Box& a_b) const;
  



  /// refinement

  ///
  /**
     Modifies this ProblemDomain by refining it by given (positive) refinement
     ratio.  The Empty ProblemDomain is not modified by this function.  \\

  */
    ProblemDomain& refine (int a_refinement_ratio);

  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain refined by given
     (positive) refinement ratio.  The Empty ProblemDomain is not modified by
     this function.  \\

  */
  friend ProblemDomain refine (const ProblemDomain& a_probdomain,
                               int   a_refinement_ratio);
  
  ///
  /**
     Modifies this ProblemDomain by refining it by given (positive) refinement
     ratio.  The Empty ProblemDomain is not modified by this function.  \\
  */
  ProblemDomain& refine (const IntVect& a_refinement_ratio);
  
  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain refined 
     by given (positive) refinement ratio.  The Empty ProblemDomain is 
     not modified by this function.  \\
     
  */
  friend ProblemDomain refine (const ProblemDomain&     a_probdomain,
                            const IntVect& a_refinement_ratio);

  /// coarsening
  
  ///
  /**
     Modifies this ProblemDomain by coarsening it by given (positive) 
     refinement ratio.  The Empty ProblemDomain is not modified by 
     this function.  \\
  */
  
  ProblemDomain& coarsen (int a_refinement_ratio);
  
  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain coarsened by given
     (positive) refinement ratio.  The Empty ProblemDomain is not modified by
     this function.  \\
     
  */
  friend ProblemDomain coarsen (const ProblemDomain& a_probdomain,
                             int        a_refinement_ratio);
  
  ///
  /**
     Modifies this ProblemDomain by coarsening by given (positive) refinement
     ratio.  The Empty ProblemDomain is not modified by this function.  \\
  */
  ProblemDomain& coarsen (const IntVect& refinement_ratio);
  
  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain coarsened 
     by given (positive) refinement ratio.  The Empty ProblemDomain is 
     not modified by this function.  \\
     
  */
  friend ProblemDomain coarsen (const ProblemDomain&  a_probdomain,
                             const IntVect& a_refinement_ratio);
  


  ///{\bf I/O Functions}
  
  ///
  /**
     Write an ASCII representation to the ostream.
     
  */
  friend std::ostream& operator<< (std::ostream&   os,
                                   const ProblemDomain& bx);

  ///
  /**
     Read from istream.
     
  */
  friend std::istream& operator>> (std::istream& is,
                                   ProblemDomain&     bx);
  
  ///
  /**
     Gives more detail than printOn.  Useful for exiting due to an
     error.  
     
  */
  void dumpOn (std::ostream& strm) const;
  
  
protected:
  friend class HDF5Handle;

  bool m_isPeriodic[SpaceDim];

  Box m_domainBox;

  ShiftIterator m_shiftIt;



};

//
// Inlines.
//
#ifndef WRAPPER

inline
ShiftIterator::ShiftIterator()
  : m_index(100), m_shift_vectors()
{
}

inline 
ShiftIterator::ShiftIterator(const ShiftIterator& a_src)
{
  m_index = a_src.m_index;
  m_shift_vectors = a_src.m_shift_vectors;
}

inline
ShiftIterator&
ShiftIterator::operator=(const ShiftIterator& a_src)
{
  m_index = a_src.m_index;
  m_shift_vectors = a_src.m_shift_vectors;
  return *this;
}

inline
IntVect
ShiftIterator::operator()() const
{
  assert(ok());
  return m_shift_vectors[m_index];
}

inline
void
ShiftIterator::operator++() 
{
  m_index++;
}

inline 
bool
ShiftIterator::ok() const
{
  return (m_index < m_shift_vectors.size());
}

inline 
void
ShiftIterator::reset()
{
  m_index = 0;
}

inline
void
ShiftIterator::begin()
{
  m_index = 0;
}

inline 
void
ShiftIterator::end()
{
  m_index = m_shift_vectors.size();
}

inline
ProblemDomain::ProblemDomain ()
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      m_isPeriodic[dir] = false;
    }
  
}

inline
ProblemDomain::ProblemDomain (const ProblemDomain& b)
  : m_domainBox(b.m_domainBox), m_shiftIt(b.m_shiftIt)
{  
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      m_isPeriodic[dir] = b.m_isPeriodic[dir];
    }
}


inline
ProblemDomain&
ProblemDomain::operator= (const ProblemDomain& b)
{  
  m_domainBox = b.m_domainBox;
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      m_isPeriodic[dir] = b.m_isPeriodic[dir];
    }
  m_shiftIt = b.m_shiftIt;
  return *this;
}

inline 
const Box& 
ProblemDomain::domainBox() const
{
  return m_domainBox;
}

inline
bool
ProblemDomain::isPeriodic(int a_dir) const
{
  return m_isPeriodic[a_dir];
}

inline
bool
ProblemDomain::isPeriodic() const
{
  return D_TERM(m_isPeriodic[0],
                || m_isPeriodic[1],
                || m_isPeriodic[2]);
}

inline
ProblemDomain& 
ProblemDomain::grow(int i)
{
  m_domainBox.grow(i);
  return *this;
}

inline 
ProblemDomain 
grow(const ProblemDomain& pd, int i)
{
  ProblemDomain newPd(pd);
  newPd.grow(i);
  return newPd;
}

inline
ProblemDomain&
ProblemDomain::grow(const IntVect& v)
{
  m_domainBox.grow(v);
  return *this;
}

inline
ProblemDomain
grow(const ProblemDomain& pd, const IntVect& v)
{
  ProblemDomain newPd(pd);
  newPd.grow(v);
  return newPd;
}

inline
ProblemDomain&
ProblemDomain::grow(int idir, int n_cell)
{
  m_domainBox.grow(idir, n_cell);
  return *this;
}

inline 
ProblemDomain&
ProblemDomain::growLo(int idir, int n_cell)
{
  m_domainBox.growLo(idir, n_cell);
  return *this;
}

inline 
ProblemDomain&
ProblemDomain::growHi(int idir, int n_cell)
{
  m_domainBox.growHi(idir, n_cell);
  return *this;
}

inline 
ShiftIterator
ProblemDomain::shiftIterator() const
{
  return m_shiftIt;
}

inline
bool
ProblemDomain::isEmpty () const
{
  return (m_domainBox.isEmpty());
}

inline
bool
ProblemDomain::contains (const IntVect& p) const
{

  // boy is this ugly!
  return ( !isEmpty() && 
           (D_TERM((m_isPeriodic[0] 
                    || (p[0]>m_domainBox.smallEnd(0) 
                        && p[0]<m_domainBox.bigEnd(0))),
                   && (m_isPeriodic[1] 
                       || (p[1]>m_domainBox.smallEnd(1) 
                           && p[1]<m_domainBox.bigEnd(1))),
                   && (m_isPeriodic[2]
                       || (p[2]>m_domainBox.smallEnd(2)
                           && p[2]<m_domainBox.bigEnd(2))))));
}
                     
inline
bool
ProblemDomain::image(IntVect& p) const
{
  if(m_domainBox.contains(p)) return true;
  if(!contains(p)) return false;

  if(m_isPeriodic[0]){
	if(p[0]<m_domainBox.smallEnd(0))    p[0]+= m_domainBox.size(0);
	else if(p[0]>m_domainBox.bigEnd(0)) p[0]-= m_domainBox.size(0);
  }
  if(m_isPeriodic[1]){
	if(p[1]<m_domainBox.smallEnd(1))    p[1]+= m_domainBox.size(1);
	else if(p[1]>m_domainBox.bigEnd(1)) p[1]-= m_domainBox.size(1);
  }
  
#if (CH_SPACEDIM == 3)
  if(m_isPeriodic[2]){
	if(p[2]<m_domainBox.smallEnd(2))    p[2]+= m_domainBox.size(2);
	else if(p[2]>m_domainBox.bigEnd(2)) p[2]-= m_domainBox.size(2);
  }
#endif
  return true;
}


inline
bool
ProblemDomain::contains (const Box& b) const
{
  // boy is this ugly!
  return ( !isEmpty() && 
           (D_TERM((m_isPeriodic[0] 
                    || (b.smallEnd(0)>=m_domainBox.smallEnd(0) 
                        && b.bigEnd(0)<=m_domainBox.bigEnd(0))),
                   && (m_isPeriodic[1] 
                       || (b.smallEnd(1)>=m_domainBox.smallEnd(1) 
                           && b.bigEnd(1)<=m_domainBox.bigEnd(1))),
                   && (m_isPeriodic[2]
                       || (b.smallEnd(2)>=m_domainBox.smallEnd(2)
                           && b.bigEnd(2)<=m_domainBox.bigEnd(2))))));
  
}

#endif /*WRAPPER*/


#endif /*CH_PROBLEM_DOMAIN_H*/
