/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// do not include memtrack.H here, since we don't want to
// track the allocation tracker itself.
//
//  addendum.  ha ha , my tracker outfoxed me anyways.  since
//  our heavy use allocators have their own tracking mechanisms
//  (BaseFab, Pool, Vector) I end up tracking some of the trackers
//  memory anyways.  I'm aware of it all, just have to be careful
//  about things.  bvs

#ifdef ENABLE_MEMORY_TRACKING

#include <list>
#include <iostream>
#include <csignal>

extern "C" {
#include <malloc.h>
}

#include "Arena.H"
#include "parstream.H"
#include "Vector.H"
#include "Pool.H"
#include "Copier.H"
#include "IntVectSet.H"
#include "MayDay.H"

long int BYTES_PER_MEG = 1024*1024;

using std::list;
using std::cout;
using std::endl;

void UnfreedMemory();

// function to register for listening to signal SIGUSR1 to
// provide a snapshot of the current memory layout
void dumpmemorymap(int sig)
{
  UnfreedMemory(); // act on signal
#ifndef CH_IRIX64
  signal(sig, dumpmemorymap); // reset self for future requests
#endif
}

void dumpmemoryatexit()
{
  Copier::s_motionItemPool.clear();
  IntVectSet::clearStaticMemory();
  /*  this code was from my mad debugging of the static order
      initialization bug in the memory tracking code that was
      making the optimized code behave differently than the 
      debug code, because static initialization order is not
      preserved during an optimizaed compilation.  bvs

  incr::iterator i;
  lst::iterator  L;
  for(i=vectorIncr_->begin(); i!= vectorIncr_->end(); i++)
    {
      cout<<i->first<<"  ";
      for(L=i->second->begin(); L!= i->second->end(); L++)
        cout << " "<<*L;
      cout<<"\n";
    }
  */ 
  UnfreedMemory();
//  delete  Arena::arenaList_;
//  delete  Pool::poolList_;
//  delete  vectorList_;
//  delete  vectorIncr_;
//  delete  Arena::s_Arena;
}

void dumpmemoryabort(int sig)
{
  UnfreedMemory();
  signal(sig, SIG_DFL);
  raise(sig);
}

int registerMemorySignals()
{
#ifndef MPI
  atexit(dumpmemoryatexit);
#ifndef CH_IRIX64
  signal(SIGABRT, dumpmemoryabort);
  signal(SIGUSR2, dumpmemorymap);
#endif
#endif
  return 1;
}

static int registerAbortHandler = registerMemorySignals();


VectorList* vectorList_ = NULL;
incr*       vectorIncr_ = NULL;

typedef struct{
  void*  address;
  size_t size;
} chunk ;

typedef struct {
  std::list<chunk*>    chunks;
  char            file[64];
  int             line;
  long            total;
  long            peak;
  bool            mlloc;
} ALLOC_INFO;

Pool chunkPool(sizeof(chunk), "memtrack:chunk");
Pool infoPool(sizeof(ALLOC_INFO), "memtrack:locations"); 

typedef std::list<ALLOC_INFO*> AllocList;

AllocList *allocList;
unsigned long grandTotal = 0;
unsigned long grandPeak  = 0;

bool trackingOn = true;

void memTrackingOn()
{
  trackingOn = true;
}

void memtrackingOff()
{
  trackingOn = false;
}


void AddTrack(void* addr, size_t asize, const char *fname, int lnum,
              bool mlloc)
{
  if(!allocList)
  {
    allocList = new(AllocList);
  }
  if(!trackingOn) return;
  if(addr==NULL) return;

  chunk* tmp = (chunk*)chunkPool.getPtr();
  tmp->address=addr;
  tmp->size=asize;

  AllocList::iterator i;
  for(i = allocList->begin(); i != allocList->end(); i++)
    {
      ALLOC_INFO& info = *(*i);
      if(info.line == lnum)
        if(strncmp(fname, info.file, strlen(fname))==0)
          {
            info.chunks.insert(info.chunks.begin(), tmp);
            info.total += asize;
            if(info.total > info.peak) info.peak=info.total;
            grandTotal += asize;
            if(grandTotal > grandPeak) grandPeak = grandTotal;
            return;
          }
    }
  ALLOC_INFO* info;
  info = (ALLOC_INFO*)infoPool.getPtr();
  new (info) ALLOC_INFO;
  info->chunks.insert(info->chunks.begin(), tmp);
  info->total = asize;
  info->peak  = asize;
  strncpy(info->file, fname, 63);
  info->line = lnum;
  info->mlloc = mlloc;
  allocList->insert(allocList->begin(), info);

}

void RemoveTrack(void* addr, bool mlloc)
{
  if(!trackingOn) return;
  if(addr==NULL || !allocList) return;

  AllocList::iterator i;
  list<chunk*>::iterator c;
  for(i = allocList->begin(); i != allocList->end(); i++)
    {
      for(c=(*i)->chunks.begin(); c!= (*i)->chunks.end(); c++)
        {
          if((*c)->address == addr)
            {
              bool m = (*i)->mlloc;
              if(mlloc != m)
                {
                  if(!m)
                    MayDay::Error("memory allocated with operator NEW returned using free");
                  else
                    MayDay::Error("memory allocated with malloc-family returned using DELETE");
                }

              (*i)->chunks.remove(*c);
              (*i)->total -= (*c)->size;
              chunkPool.returnPtr(*c);
              // if((*i)->chunks.empty()){
              //    allocList->remove(*i);
              //    infoPool.returnPtr((*i));
              // }
              return;
            }
        }
    }   
  MayDay::Error("free/DELETE called on un-malloced/NEW'd pointer, or double freed");
}



void ReportUnfreedMemory(ostream& os)
{
  AllocList::iterator  i;
  PoolList::iterator   p;
  ArenaList::iterator  a;
  VectorList::iterator v;

  long int totalSize = 0;
  long int peak = 0;

  if(trackingOn && allocList){
    for(i = allocList->begin(); i != allocList->end(); i++)
    {
      peak += (*i)->peak;
      if( (*i)->total != 0)
      {
        totalSize += (*i)->total;
        //os<< "-----------------------------------------------------------\n";
        os << (*i)->file<<":"<<(*i)->line<<":  "<< (*i)->total<<  "\n";
      }
    }
    if(totalSize != 0)
    {
      os << "Total Unfreed from new,malloc, etc: "<<totalSize<<"  bytes\n";
    }
  } 

  if(vectorList_ != NULL)
  {
    //  os<<"---------------------------------\n";
    //  os<<"Vector-based memory allocation\n";
    for(v = vectorList_->begin(); v!=vectorList_->end(); v++)
      {
        long int* bytes = (v->second.first);

        if(*bytes != 0 && *bytes != 8)
        {
          os<<"Vector "<<v->first<<": "<<*bytes<<" bytes ("
            <<(*bytes)/BYTES_PER_MEG<<" Mb)\n";
          totalSize+=*bytes;
        }
        peak += *(v->second.second);
        // probably wondering about my silly "8" check.  workaround for
        // now since I can't clear chunkPool and infoPool unless this
        // is an atexit call, since they need to stay fresh in case this
        // is a transient run call to memusage.  will think up a better
        // way for this at another time.  Can't kill myself for 8 bytes
        //   bvs.
      }
  }

  //  os<<"---------------------------------\n";
  //  os<<"Pool-based memory allocation\n";
  for(p = Pool::poolList_->begin(); p!=Pool::poolList_->end(); p++)
    {
      peak+= (*p)->memUsage();

      if(*p == &chunkPool || *p == &infoPool)
      {
        // os<<(*p)->name_<<": "<<(*p)->memUsage()<<" bytes ("
        //   <<(*p)->memUsage()/BYTES_PER_MEG<<" Mb) (not added to total)\n";
      }
      else if ((*p)->memUsage() != 0)
      {
        os<<(*p)->name_<<": "<<(*p)->memUsage()<<" bytes ("
          <<(*p)->memUsage()/BYTES_PER_MEG<<" Mb)\n";
        totalSize+= (*p)->memUsage();
      }
    }

  if(Arena::arenaList_ != NULL)
  {
    // os<<"---------------------------------\n";
    // os<<"BaseFab-based memory allocation\n";
  for(a = Arena::arenaList_->begin(); a!=Arena::arenaList_->end(); a++)
    {
      peak+=(*a)->peak;
      if((*a)->bytes != 0)
      {
        os<<"BaseFab "<<(*a)->name_<<": "<<(*a)->bytes<<" bytes ("
          <<(*a)->bytes/BYTES_PER_MEG<<" Mb)\n";
        totalSize+= (*a)->bytes;
      }
    }
  }

  peak+= BitSet::peak + IntVectSet::peakcount*sizeof(IntVectSet);

  if(BitSet::bytes != 0)
  {
    //os<<"---------------------------------\n";
    os<<" BitSet allocation "<< BitSet::bytes<<" bytes ("
      <<BitSet::bytes/BYTES_PER_MEG<<" Mb)\n";
    totalSize += BitSet::bytes;
  }
  if(IntVectSet::count != 0)
  {
    os<<" IntVectSet allocation "<< IntVectSet::count*sizeof(IntVectSet)
      <<" bytes ("
      <<IntVectSet::count*sizeof(IntVectSet)/BYTES_PER_MEG<<" Mb)\n";
    totalSize += IntVectSet::count*sizeof(IntVectSet);
  }
  if(totalSize != 0)
  {
    os<<"---------------------------------\n";  
    os << "Total Unfreed : "<<totalSize<<" bytes ("
       <<totalSize/BYTES_PER_MEG<<" Mb)\n";
  }
  os<<"peak memory usage: "<<peak<<" bytes ("
    <<peak/BYTES_PER_MEG<<" Mb)\n";
  os<<std::endl;

}
void UnfreedMemory()
{
  ReportUnfreedMemory(pout());
}

void ReportMemoryUsage(ostream& os)
{
  ReportUnfreedMemory(os);
}

void MemoryUsage()
{
  ReportUnfreedMemory(pout());
}

void overallMemoryUsage(long int& currentTotal, long int& peak)
{
  currentTotal = 0;
  peak = 0;
  AllocList::iterator  i;
  PoolList::iterator   p;
  ArenaList::iterator  a;
  VectorList::iterator v;

  if(trackingOn && allocList)
  {
    for(i = allocList->begin(); i != allocList->end(); i++)
    {
      if( (*i)->total != 0)
      {
        currentTotal += (*i)->total;
        peak += (*i)->peak;
      }
    }
  } 

  if(vectorList_ != NULL)
  {
    for(v = vectorList_->begin(); v!=vectorList_->end(); v++)
      {
        long int* bytes = (v->second.first);
        currentTotal += *bytes;
        peak += *(v->second.second); // little trick might break some day.  bvs
      }
  }

  for(p = Pool::poolList_->begin(); p!=Pool::poolList_->end(); p++)
    {
      if(*p == &chunkPool || *p == &infoPool)
      {
        // don't report on the tracker itself.
      }
      else
      {
        currentTotal += (*p)->memUsage();
        peak+= (*p)->memUsage();
      }
    }

  if(Arena::arenaList_ != NULL){
  for(a = Arena::arenaList_->begin(); a!=Arena::arenaList_->end(); a++)
    {
      currentTotal += (*a)->bytes;
      peak += (*a)->peak;
    }
  }

  currentTotal += BitSet::bytes;
  peak += BitSet::peak;

  currentTotal += IntVectSet::count*sizeof(IntVectSet);
  peak += IntVectSet::peakcount*sizeof(IntVectSet);

}


//extra piece of Brian trickier to make sure that mallocp etc.
// do not become recursive and ridiculous.  now we need access
// to the real memory functions.

#undef malloc
#undef realloc
#undef calloc
#undef free


void* mallocp(size_t size, const char* file, int line)
{
  void *ptr = (void *)malloc(size);
  AddTrack(ptr, size, file, line, true);
  return(ptr);
}

void* reallocp(void* p, size_t size, const char* file, int line)
{
  RemoveTrack(p, true);
  void *ptr = (void *)realloc(p,size);
  AddTrack(ptr, size, file, line, true);
  return(ptr);
}

void *callocp(size_t nelem, size_t elsize, const char* file, int line)
{
  void *ptr = (void *)calloc(nelem, elsize);
  AddTrack(ptr, elsize*nelem, file, line, true);
  return(ptr);
}

void freep(void* p)
{
  RemoveTrack(p, true);
  free(p);
}

// void * operator new(size_t size, char const *file, int line)
// {
//   void *ptr = (void *)malloc(size);
//   AddTrack(ptr, size, file, line, false);
//   return(ptr);
// }

// void operator delete(void *p) throw()
// {
//   RemoveTrack(p, false);
//   free(p);
// }

// void * operator new[](size_t size, char const *file, int line)
// {
//   void *ptr = (void *)malloc(size);
//   AddTrack(ptr, size, file, line, false);
//   return(ptr);
// }

// void operator delete[](void *p) throw()
// {
//   RemoveTrack(p, false);
//   free(p);
// }

#include "SPMD.H"


#ifdef MPI

void mpierrorfunction(MPI_Comm* comm, int* error, ...)
{
  char string[MPI_MAX_ERROR_STRING];
  int resultlen;
  MPI_Error_string(*error, string, &resultlen);
  cout<< string<<"\n";
  dumpmemoryatexit();
  // it would be nice to have a stack dump at this point also.
  MPI_Abort(*comm, *error);
}

void setChomboMPIErrorHandler()
{
  MPI_Errhandler handler;
  MPI_Errhandler_create(mpierrorfunction, &handler);
  MPI_Errhandler_set(Chombo_MPI::comm, handler);
}
#endif

#else    // on ENABLE_MEMORY_TRACKING
// make some stub functions

void dumpmemoryatexit(){}
void setChomboMPIErrorHandler(){}




#endif  // on ENABLE_MEMORY_TRACKING


