/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LevelFluxRegister.H"
#include "LoadBalance.H"
#include "DebugDump.H"
#include "LayoutIterator.H"

/*****************/
/*****************/
void 
setDefaults( int& nref, 
             Vector<Box>& coarboxes,
             Vector<Box>& fineboxes, 
             Box& domc, Box& domf);

extern int fluxRegTest(void);
void parseTestOptions(int argc ,char* argv[]) ;
/*****************/
/*****************/

/// Global variables for handling output:
static const char *pgmname = "fluxRegisterTest" ;
static const char *indent = "   " ,*indent2 = "      " ;
static bool verbose = true ;

/*****************/
/*****************/

int
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  int eekflag = fluxRegTest();
  if(eekflag == 0)
    {
      cout << indent << pgmname
           << ": test passed." << endl;
    }
  else
    {
      cout << indent << pgmname
           << ": test FAILED with error code " 
           << eekflag << endl;
    }
#ifdef MPI
  MPI_Finalize();
#endif
  return eekflag;
}

/*****************/
/*****************/

int fluxRegTest()
{
  int nref;
  Vector<Box> fineboxes;
  Vector<Box> coarboxes;
  Box domf, domc;

  //set fine grid
  setDefaults(nref, coarboxes,  fineboxes, domc, domf);
  int retflag = 0;

  Interval interv(0,0);
  //loop through directions
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      //set up coarse and fine grids 
      DisjointBoxLayout dblFineCell,dblCoarCell;
      Vector<int> procAssignCoar(coarboxes.size(), 0);
      Vector<int> procAssignFine(fineboxes.size(), 0);
      LoadBalance(procAssignCoar, coarboxes);
      LoadBalance(procAssignFine, fineboxes);
      //only want edge centered data on the interior
      Box edgeIntCoar = surroundingNodes(domc, idir);
      edgeIntCoar.grow(idir, -1);
      Box edgeIntFine = surroundingNodes(domf, idir);
      edgeIntFine.grow(idir, -1);

      dblCoarCell.define(coarboxes, procAssignCoar);
      dblFineCell.define(fineboxes, procAssignFine);

      dblCoarCell.close();
      dblFineCell.close();
      if(verbose)
        {
          cout << indent2 << pgmname
               << ": dblCoarCell is" << endl;
          dumpDBL(&dblCoarCell);
          cout << indent2 << pgmname
               << ": dblFineCell is" << endl;
          dumpDBL(&dblFineCell);
        }

      LevelData<FArrayBox> coarData(dblCoarCell, 1);
      LevelData<FArrayBox> fineData(dblFineCell, 1);

      DataIterator coarIt = coarData.dataIterator();
      DataIterator fineIt = fineData.dataIterator();

      LevelFluxRegister fluxReg(dblFineCell, 
                                dblCoarCell, 
                                domf, nref, 
                                1);

      //set data and flux registers to zero
      for(coarIt.reset(); coarIt.ok(); ++coarIt)
        coarData[coarIt()].setVal(0.);

      fluxReg.setToZero();

      if(verbose)
        {
          cout << indent2 << pgmname
               << ": after setting to zero, the fluxReg in dir " << idir << " is" << endl;
          fluxReg.dumpLoCoar(idir);
          fluxReg.dumpHiCoar(idir);
          fluxReg.dumpLoFine(idir);
          fluxReg.dumpHiFine(idir);
        }

      //increment and decrement 
      //flux registers with equal size fluxes
      Real scale = 1.0;
      Real fluxVal = 4.77;
      for(coarIt.reset(); coarIt.ok(); ++coarIt)
        {
          const Box&  cellBoxCoar = dblCoarCell.get(coarIt());
          Box edgeBoxCoar = surroundingNodes(cellBoxCoar, idir);
          edgeBoxCoar &= edgeIntCoar;
          FArrayBox edgeFlux(edgeBoxCoar,1);
          edgeFlux.setVal(fluxVal);
          DataIndex dataIndGlo = coarIt();
          fluxReg.incrementCoarse(edgeFlux, scale,
                                  dataIndGlo, interv, interv, idir);
        }
      if(verbose)
        {
          cout << indent2 << pgmname
               << ": after coarse increment, the fluxReg in dir " << idir << " is" << endl;
          fluxReg.dumpLoCoar(idir);
          fluxReg.dumpHiCoar(idir);
          fluxReg.dumpLoFine(idir);
          fluxReg.dumpHiFine(idir);
        }



      for(fineIt.reset(); fineIt.ok(); ++fineIt)
        {
          const Box&  cellBoxFine = dblFineCell.get(fineIt());
          Box edgeBoxFine = surroundingNodes(cellBoxFine, idir);
          edgeBoxFine &= edgeIntFine;
          FArrayBox edgeFlux(edgeBoxFine,1);
          edgeFlux.setVal(fluxVal);
          SideIterator sit;
          DataIndex dataIndGlo = fineIt();
          for(sit.reset(); sit.ok(); ++sit)
            {
              fluxReg.incrementFine(edgeFlux, scale,
                                    dataIndGlo,  interv, interv, idir, sit());
            }
        }
      if(verbose)
        {
          cout << indent2 << pgmname
               << ": after fine increment, the fluxReg in dir " << idir << " is" << endl;
          fluxReg.dumpLoCoar(idir);
          fluxReg.dumpHiCoar(idir);
          fluxReg.dumpLoFine(idir);
          fluxReg.dumpHiFine(idir);
        }


      //reflux what ought to be zero into zero and the result should be zero
      fluxReg.reflux(coarData, scale);
      if(verbose)
        {
          cout << indent2 << pgmname
               << ": after reflux, the fluxReg in dir " << idir << " is" << endl;
          fluxReg.dumpLoCoar(idir);
          fluxReg.dumpHiCoar(idir);
          fluxReg.dumpLoFine(idir);
          fluxReg.dumpHiFine(idir);
        }


      DataIterator datIt = coarData.dataIterator();
      for(datIt.reset(); datIt.ok(); ++datIt)
        {
          const FArrayBox& data = coarData[datIt()];
          Real rmax = abs(data.max());
          Real rmin = abs(data.min());
          if((rmax > 1.0e-10)||(rmin > 1.0e-10))
            {
              cout << indent << pgmname
                   << ": fluxRegister failed the test at idir = "
                   << idir << endl;
              retflag = 1;
            }
        }
    }
  return retflag;
}

/*****************/
void 
setDefaults( int& nref, 
             Vector<Box>& coarboxes, 
             Vector<Box>& fineboxes, 
             Box& domc, Box& domf)
{
  nref = 2;
  int nc = 16;
  int nc16 = nc/2;
  int nc8 = nc/4; 
  int nc24 = 3*nc/4; 
  int nc12 = 3*nc/8; 
  int nc20 = 5*nc/8; 

  IntVect ivclo = IntVect::TheZeroVector();

  fineboxes.resize(0);
#if(CH_SPACEDIM ==2)

  IntVect ivchi(nc-1, nc-1);
  Box boxf1(IntVect(0 ,nc8), IntVect(nc16-1,nc24));
  Box boxf2(IntVect(nc16,nc12), IntVect(nc24-1,nc20-1));
  Box boxf3(IntVect(nc8 ,0 ), IntVect(nc20-1,nc8-1));
  Box boxf4(IntVect(nc20,nc8 ), IntVect(nc-1,nc12-1));

#elif(CH_SPACEDIM==3)
  IntVect ivchi(nc-1, nc-1, nc-1);
  Box boxf1(IntVect(0 ,nc8,0), IntVect(nc16-1,nc24,nc-1));
  Box boxf2(IntVect(nc16,nc12,0), IntVect(nc24-1,nc20-1,nc-1));
  Box boxf3(IntVect(nc8 ,0 ,0), IntVect(nc20-1,nc8-1,nc-1));
  Box boxf4(IntVect(nc20,nc8,0 ), IntVect(nc-1,nc12-1,nc-1));
#else
  error bogus spacedim;
#endif
  boxf1.refine(2);
  boxf2.refine(2);
  boxf3.refine(2);
  boxf4.refine(2);

  domc = Box(ivclo, ivchi);
  int ichop = domc.size(0)/2;
  Box bc1 = domc;
  Box bc2 = bc1.chop(0, ichop);
  Box bc3 = bc2.chop(1, ichop);
  coarboxes.push_back(bc1);
  coarboxes.push_back(bc2);
  coarboxes.push_back(bc3);

  fineboxes.push_back(boxf1);
  fineboxes.push_back(boxf2);
  fineboxes.push_back(boxf3);
  fineboxes.push_back(boxf4);
  domf = refine(domc,nref);
}


///
// Parse the standard test options (-v -q) out of the command line.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
        }
    }
  return ;
}
