/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "REAL.H"
#include "Box.H"
#include "Vector.H"
#include "IntVectSet.H"
#include "BoxIterator.H"
#include "TreeIntVectSet.H"
#include "ProblemDomain.H"
#include "MayDay.H"
#ifdef MPI
#include <mpi.h>
#endif

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

int
testIntVectSet();


/// Global variables for handling output:
static const char *pgmname = "testIntVectSet" ;
static const char *indent = "   ", *indent2 = "      " ; 
static bool verbose = true ;


/// Code:

int 
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions( argc ,argv ) ;

  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  ///
  // Run the tests
  ///
  int ret = testIntVectSet() ;
#ifdef MPI
  MPI_Finalize();
#endif
  return ret;
}


int
testIntVectSet()
{
  bool eekflag = false;

  {
    // dan's silly test
    Box testBox(IntVect::Zero, 3*IntVect::Unit);
    ProblemDomain testDom(testBox);
    for (int dir=0; dir<SpaceDim; dir++) 
      {
        testDom.setPeriodic(dir,true);
      }

    int nestingRadius = 1;
    IntVectSet testIVS;
    testIVS |= testBox;
    testIVS.nestingRegion(nestingRadius, testDom);

	TreeIntVectSet tivs;
	Box set(3*IntVect::Unit, 4*IntVect::Unit);
	tivs |= set;

	for(BoxIterator bit(set); bit.ok(); ++bit)
	  {
		IntVect image = bit();
		if(!testDom.image(image)){
		  MayDay::Error(" A point managed to fall outside fully periodic");
		}
		if(!testBox.contains(image)){
		  MayDay::Error("image point falls outside domain box");
		}
	  }
	if(tivs.numPts() != D_TERM(2,*2,*2)){
	  MayDay::Error("points went missing on ProblemDomain intersection");
	}
	
	cout << indent << pgmname
		 << ": IntVectSet passed periodic ProblemDomain test" 
		 << endl;
	
  }
        

  {
    IntVectSet ivs;
    IntVect a(-1*IntVect::Unit), b(0*IntVect::Unit);
    ivs |= a;
    if(!ivs.contains(a))
      {
        cout << indent << pgmname
             << ": failed -1 -1 -1 add test" 
             << endl;
        eekflag = true;
      }
    if(ivs.numPts() != 1)
      {
        cout << indent << pgmname
             << ": failed single IntVect count test" 
             << endl;
        eekflag = true;
      }
    ivs-=a;
    if(ivs.contains(a))
      {
        cout << indent << pgmname
             << ": failed [-1 -1 -1] remove test" 
             << endl;
        eekflag = true;
      }
    if(ivs.numPts() != 0 || !ivs.isEmpty())
      {
        cout << indent << pgmname
             << ": failed single IntVect removal count test" 
             << endl;
        eekflag = true;
      }
    ivs|=a;
    ivs|=b;
    if(!ivs.contains(a) || !ivs.contains(b))
      {
        cout << indent << pgmname
             << ": failed [-1 -1 -1] [0,0,0] add test" 
             << endl;
        eekflag = true;
      }
    if(ivs.numPts() != 2)
      {
        cout << indent << pgmname
             << ": failed two IntVect count test" 
             << endl;
        eekflag = true;
      }


  }

#ifndef CH_AIX  // not working on seaborg with xlC -- but don't
                // know how to be more specific with ifdef's right here.  (ndk)
  dumpTree(NULL); // does nothing, but makes sure dumpTree is avaiable for
#endif
  // debugging work

  Box b1(12 * IntVect::TheUnitVector(), 
         31 * IntVect::TheUnitVector());
  Box b2(32 * IntVect::TheUnitVector(), 
         53 * IntVect::TheUnitVector());



  {
    // add, remove box.  success = empty IntVectSet.
    IntVectSet ivs;
    ivs |= b1;
    ivs |= b2;

    ivs -= b1;
    ivs -= b2;

    if (ivs.isEmpty()) 
      {
        cout << indent << pgmname
             << ": IntVectSet passed the add/remove Box test" 
             << endl;
      }
    else
      {
        cout << indent << pgmname 
             << ": IntVectSet FAILED the add/remove Box test" 
             << endl;
        eekflag = true;
      }
  }

  // intersect with box, remove box.  success = empty IntVectSet.
  {
    IntVectSet ivs;
    ivs |= b1;
    ivs |= b2;

    ivs &= b1;
    ivs -= b1;

    if (ivs.isEmpty()) 
      {
        cout << indent << pgmname
             << ": IntVectSet passed the intersect/remove Box test" 
             << endl;
      }
    else
      {
        cout << indent << pgmname
             << ": IntVectSet FAILED the intersect/remove Box test" 
             << endl;
        eekflag = true;
      }
  }

  // add box.  success = every IntVect is in the IVS.
  {
    IntVectSet ivs;
    ivs |= b1;
    BoxIterator bit(b1);
    Vector<IntVect> iv_fail;
    for (bit.begin(); bit.ok(); bit.next())
      {
        IntVect iv = bit();
        if (!ivs.contains(iv))
          {
            iv_fail.push_back(iv);
          }
      }
    if (iv_fail.size() == 0)
      {
        cout << indent << pgmname
             << ": IntVectSet passed the every IntVect in the Box test"
             << endl;
      }
    else
      {
        cout << indent << pgmname
             << ": IntVectSet FAILED the every IntVect in the Box test"
             << endl;
        eekflag = true;
      }
  }    

  // intersect disjoint IntVectSets.  success = empty IntVectSet.
  {
    IntVectSet ivs1(b1);
    IntVectSet ivs2(b2);
    IntVectSet ivs_intersect = ivs1 & ivs2;

    if (ivs_intersect.isEmpty())
      {
        cout << indent << pgmname
             << ": IntVectSet passed the disjoint intersection test"
             << endl;
      }
    else
      {
        cout << indent << pgmname
             << ": IntVectSet FAILED the disjoint intersection test"
             << endl;
        eekflag = true;
      }
  }
  // remove  one intvect from box. 
  {
    IntVectSet ivs(b1);
    IntVect iv = (b1.smallEnd() + b1.bigEnd() ) / 2;
    ivs -= iv;
    BoxIterator bit(b1);
    Vector<IntVect> iv_fail;
    for (bit.begin(); bit.ok(); bit.next())
      {
        if(iv == bit()){;}
        else if (!ivs.contains(bit()))
          {
            iv_fail.push_back(bit());
          }
      }
    if (iv_fail.size() == 0 && !ivs.contains(iv))
      {
        cout << indent << pgmname
             << ": IntVectSet passed the remove IntVect test"
             << endl;
      }
    else
      {
        cout << indent << pgmname
             << ": IntVectSet FAILED the remove IntVect test"
             << endl;
        eekflag = true;
      }
  }

  // remove and add one intvect from box.  success = entire box is in IntVectSet.
  {
    IntVectSet ivs(b1);
    IntVect iv = (b1.smallEnd() + b1.bigEnd() ) / 2;
    ivs -= iv;
    ivs |= iv;
    BoxIterator bit(b1);
    Vector<IntVect> iv_fail;
    for (bit.begin(); bit.ok(); bit.next())
      {
        IntVect iv = bit();
        if (!ivs.contains(iv))
          {
            iv_fail.push_back(iv);
          }
      }
    if (iv_fail.size() == 0)
      {
        cout << indent << pgmname
             << ": IntVectSet passed the remove/add IntVect test"
             << endl;
      }
    else
      {
        cout << indent << pgmname
             << ": IntVectSet FAILED the remove/add IntVect test"
             << endl;
        eekflag = true;
      }
  }

  {
    Box b1(IntVect(D_DECL(13,5,-1)), 
           31 * IntVect::TheUnitVector());
    Box b2(32 * IntVect::TheUnitVector(), 
           53 * IntVect::TheUnitVector());
    Box b3(14 * IntVect::TheUnitVector(), 
           53 * IntVect::TheUnitVector());


    int numpts = b2.numPts()+b3.numPts() + b1.numPts() -(b2&b3).numPts() -
      (b2&b1).numPts() - (b1&b3).numPts();

    IntVectSet ivs2(b3);

    ivs2 |= b2;

    ivs2 |= b1;

    cout <<endl;
    int count=0;

    IntVectSet otherSet(ivs2);

    IVSIterator it(otherSet), it2(ivs2);

    for(it.begin(); it.ok(); ++it)
      {
        if(!(b2.contains(it()) || b1.contains(it()) || b3.contains(it())))
          {
            if( verbose )
              {
                cout << indent2 << "union point falls outside region: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
        count++;
      }
    if(count != numpts)
      {
        cout << indent2 << "count != numpts from union ";
        cout << pgmname << endl ;
        eekflag = true;
      }

    IntVect shifter(D_DECL(1,-2,3));
    ivs2.shift(shifter);
    b1.shift(shifter);
    b2.shift(shifter);
    b3.shift(shifter);
    for(it2.begin(); it2.ok(); ++it2)
      {
        if(!(b2.contains(it2()) || b1.contains(it2()) || b3.contains(it2())))
          {
            if( verbose )
              {
                cout << indent2 << "union point falls outside region: shift";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
      }

    b1.refine(4);
    b2.refine(4);
    b3.refine(4);
    ivs2.refine(4);
    numpts = b2.numPts()+b3.numPts() + b1.numPts() -(b2&b3).numPts() -
      (b2&b1).numPts() - (b1&b3).numPts();
    count = 0;
    for(it2.begin(); it2.ok(); ++it2)
      {
        if(!(b2.contains(it2()) || b1.contains(it2()) || b3.contains(it2())))
          {
            if( verbose )
              {
                cout << indent2 << "union point falls outside region:refine ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
        count++;
      }
    if(count != numpts)
      {
        cout << indent2 << "count != numpts from union ";
        cout << pgmname << endl ;
        eekflag = true;
      }
  }

  {
    IntVect a(D_DECL(-2,-1,0)), b(D_DECL(0,1,1)), c(D_DECL(14,3,4)), 
      d(D_DECL(16,6,5));
    Box b1(a,c);
    Box b2(b,d);
    Box b3(b,c);
    IntVectSet ivs;
    ivs|= b1;
    ivs&= b2;


    long count = 0;
    for(IVSIterator it(ivs); it.ok(); ++it)
      {
        if(!b3.contains(it()))
          {
            if( verbose )
              {
                cout << indent2 << "intersection point falls outside region:";
                cout << pgmname << endl ;
              }
            return 2;
          }     
        count++;
      }
    if(count != b3.numPts())
      {
        cout << indent2 << "count != numpts from intersection ";
        cout << pgmname << endl ;
        eekflag = true;
      }

    BoxIterator bit(b1);
    for(bit.begin(); bit.ok(); ++bit)
      {
        if(b2.contains(bit()))
          {
            if(!ivs.contains(bit()))
              {
                if( verbose )
                  {
                    cout << indent2 << "intersection point missing";
                    cout << pgmname << endl ;
                    eekflag = true;
                  }     
              }
          }
        else
          {
            if(ivs.contains(bit()))
              {
                if( verbose )
                  {
                    cout << indent2 << "extraneous intersection point";
                    cout << pgmname << endl ;
                    eekflag = true;
                  }     
              }
          }
      }
  }


  {
    IntVect a(D_DECL(-2,-1,0)), b(D_DECL(0,1,1)), c(D_DECL(14,3,4)), 
      d(D_DECL(16,6,5));
    Box b1(a,c);
    Box b2(b,d);
    Box b3(b,c);
    TreeIntVectSet ivs, ivs2;
    ivs2  |= b1;
    ivs   |= b2;

    ivs&=ivs2;

    long count = 0;
    for(TreeIntVectSetIterator it(ivs); it.ok(); ++it)
      {
        if(!b3.contains(it()))
          {
            if( verbose )
              {
                cout << indent2 << "intersection point falls outside region:";
                cout << pgmname << endl ;
              }
            return 2;
          }     
        count++;
      }
    if(count != b3.numPts())
      {
        cout << indent2 << "count != numpts from intersection ";
        cout << pgmname << endl ;
        eekflag = true;
      }

    BoxIterator bit(b1);
    for(bit.begin(); bit.ok(); ++bit)
      {
        if(b2.contains(bit()))
          {
            if(!ivs.contains(bit()))
              {
                if( verbose )
                  {
                    cout << indent2 << "intersection point missing";
                    cout << pgmname << endl ;
                    eekflag = true;
                  }     
              }
          }
        else
          {
            if(ivs.contains(bit()))
              {
                if( verbose )
                  {
                    cout << indent2 << "extraneous intersection point";
                    cout << pgmname << endl ;
                    eekflag = true;
                  }     
              }
          }
      }
  }

  // make sure no strange descuctor behaviour floats around....
  {
    Box b1(12 * IntVect::TheUnitVector(), 
           31 * IntVect::TheUnitVector());
    Box b2(32 * IntVect::TheUnitVector(), 
           53 * IntVect::TheUnitVector());
    Box b3(14 * IntVect::TheUnitVector(), 
           51 * IntVect::TheUnitVector());


    IntVectSet mask1(b3);

    mask1 -= b2;
    mask1 -= b1;

    IntVectSet mask3(mask1);
    IntVectSet mask4 = mask3; //ensure copy and assignment of dense works

    for(IVSIterator it(mask3), it4(mask4); it.ok(); ++it, ++it4)
      {
        if(b2.contains(it()) || b1.contains(it()))
          {
            if( verbose )
              {
                cout << indent2 << "subtraction left in bad data: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }

        if(!b3.contains(it()))
          {
            if( verbose )
              {
                cout << indent2 << "IntVect beyond domain: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
        if(it() != it4())
          {
            if( verbose )
              {
                cout << indent2 << "IntVectSet copy/assign problem: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
      }

    IntVectSet mask2(b3);
    mask2 &= b1;

    for(IVSIterator it(mask2); it.ok(); ++it)
      {
        if(!(b3.contains(it()) && b1.contains(it())))
          {
            if( verbose )
              {
                cout << indent2 << "intersection contains extra data: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
      }
    for(BoxIterator bit(b1&b3); bit.ok(); ++bit)
      {
        if(!mask2.contains(bit()))
          {
            if( verbose )
              {
                cout << indent2 << "intersection missed data: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
      }
  }

  {
    Box b1(IntVect(D_DECL(13,5,-1)), 
           31 * IntVect::TheUnitVector());
    Box b2(32 * IntVect::TheUnitVector(), 
           53 * IntVect::TheUnitVector());
    Box b3(14 * IntVect::TheUnitVector(), 
           51 * IntVect::TheUnitVector());

    IntVectSet ivs2;

    ivs2 |= b3;

    ivs2 |= b2;

    ivs2 |= b1;

    b1.coarsen(2); b2.coarsen(2); b3.coarsen(2);
    ivs2.coarsen(2);

    int count = 0;
    int numpts = b2.numPts()+b3.numPts() + b1.numPts() -(b2&b3).numPts() -
      (b2&b1).numPts() - (b1&b3).numPts();

    IVSIterator it(ivs2);
    for(it.begin(); it.ok(); ++it)
      {
        if(!(b2.contains(it()) || b1.contains(it()) || b3.contains(it())))
          {
            if( verbose )
              {
                cout << indent2 << "box.coarsen != ivs.coarsen: point falls outside region: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
        count++;
      }
    if(count != numpts)
      {
        cout << indent2 << "box.coarsen != ivs.coarsen: count != numpts from union ";
        cout << pgmname << endl ;
        eekflag = true;
      }

  }
   {
    Box b1(IntVect(D_DECL(13,5,-1)), 
           31 * IntVect::TheUnitVector());
    Box b2(32 * IntVect::TheUnitVector(), 
           53 * IntVect::TheUnitVector());
    Box b3(14 * IntVect::TheUnitVector(), 
           51 * IntVect::TheUnitVector());

    TreeIntVectSet ivs2, ivs3, ivs4;

	ivs4.clear(); 
	ivs4.clear();

    ivs2 |= b3;

    ivs2 |= b2;

    ivs3 |= b1;

        ivs2 |= ivs3;

    b1.coarsen(2); b2.coarsen(2); b3.coarsen(2);

    ivs2.coarsen(2);

    int count = 0;
    int numpts = b2.numPts()+b3.numPts() + b1.numPts() -(b2&b3).numPts() -
      (b2&b1).numPts() - (b1&b3).numPts();

    TreeIntVectSetIterator it(ivs2);
    for(it.begin(); it.ok(); ++it)
      {
        if(!(b2.contains(it()) || b1.contains(it()) || b3.contains(it())))
          {
            if( verbose )
              {
                cout << indent2 << "box.coarsen != tivs.coarsen: point falls outside region: ";
                cout << pgmname << endl ;
              }
            eekflag = true;
          }
        count++;
      }
    if(count != numpts)
      {
        cout << indent2 << "box.coarsen != tivs.coarsen: count != numpts from union ";
        cout << pgmname << endl ;
        eekflag = true;
      }
	ivs2.clear();
	ivs2.clear();
  }
   {
	 Box b1(IntVect(D_DECL(5,5,5)), 
			10 * IntVect::TheUnitVector());
	 Box b2(6 * IntVect::TheUnitVector(), 
			7 * IntVect::TheUnitVector());
	 Box b3(IntVect(D_DECL(6,8,8)), 
			IntVect(D_DECL(7,9,8)));
	 Box b4(IntVect(D_DECL(8,6,5)),
			IntVect(D_DECL(9,7,5)));
	 Box b5(IntVect(D_DECL(8,8,8)),
			IntVect(D_DECL(9,9,9)));
	 
			   

	 TreeIntVectSet ivs1, ivs2;

	 ivs1 |= b1;

	 ivs2 |= b2;
	 ivs2 |= b3;
	 ivs2 |= b4;
	 ivs2 |= b5;

	 TreeIntVectSet ivs3(ivs2);

	 ivs2 &= ivs1;
	 TreeIntVectSetIterator it(ivs3);
	 for(it.begin(); it.ok(); ++it)
	   {
		 if(!ivs2.contains(it())){
		   MayDay::Error ("TreeIntVectSet::&= operator broken");
		 }
	   }




   }
  if(eekflag) return 1;
  else        return 0;
}


///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
